# Code for accuracy of the implicit solution
import svcco
import pyvista as pv
from time import perf_counter
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from itertools import chain
from tqdm import tqdm
###########################################
# Code for Building Surface
###########################################

cu = pv.Cube(x_length=3.72,y_length=3.72,z_length=3.72).triangulate().subdivide(5)
cube = svcco.surface()
cube.set_data(cu.points,cu.point_normals)
cube.solve()
cube.build(q=4,resolution=120)
print('cube constructed')

heart = svcco.surface()
heart_points = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_points_unique.csv',delimiter=',')
heart_normals = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_normals_unique.csv',delimiter=',')
heart.set_data(heart_points,heart_normals)
heart.solve()
heart.build(q=4,resolution=120,k=2,buffer=5)
print('heart constructed')

disk = pv.Disc(r_res=20,c_res=100)
cyl  = disk.extrude([0,0,1],capping=True).triangulate()
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl)
cyl  = cyl.subdivide(2)
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl,hausd=0.005)
cyl  = cyl.compute_normals(auto_orient_normals=True,feature_angle=90)
cylinder = svcco.surface()
cylinder.set_data(cyl.points,cyl.point_normals)
cylinder.solve()
cylinder.build(q=4,k=2,resolution=120)
print('cylinder constructed')

left_gyrus   = "D:\\Tree\\Tree_8-0\\brain_testing\\FJ3801_BP58201_FMA72658_Left inferior frontal gyrus.obj"
gyrus = svcco.surface()
gyrus.load(left_gyrus)
gyrus.solve()
gyrus.build(q=4,k=2,resolution=120,buffer=5)
print('gyrus constructed')

###########################################
# Code for evaluation accuracy
###########################################

def place_bin(value,bin_num,bin_range):
    ind = None
    BINS = np.linspace(bin_range[0],bin_range[1],bin_num+1)
    for i in range(len(BINS)-1):
        if value > BINS[i] and value < BINS[i+1]:
            ind = i
            break
    return ind

def get_values(s,value_min,value_max,reps,samples):
    count = 0
    values  = []
    patches = []
    pbar = tqdm(total=reps)
    while count < reps:
        xb = s.x_range[1] - s.x_range[0]
        yb = s.y_range[1] - s.y_range[0]
        zb = s.z_range[1] - s.z_range[0]
        xp = (s.x_range[0] + np.random.random(1)*xb).item()
        yp = (s.y_range[0] + np.random.random(1)*yb).item()
        zp = (s.z_range[0] + np.random.random(1)*zb).item()
        tmp = s.DD[0]([xp,yp,zp,len(s.patches)])[0]
        v = []
        p = []
        if tmp > value_max or tmp < value_min:
            continue
        if np.isnan(tmp):
            continue
        sample = np.linspace(2,len(s.patches),samples,dtype=int)
        for i in sample:
            v.append(s.DD[0]([xp,yp,zp,i]).tolist())
            p.append(i)
        v = np.array(v)
        p = np.array(p)
        v = abs(v - v[-1])
        v = v/max(v)
        p = p/max(p)
        values.append(v.tolist())
        patches.append(p.tolist())
        count += 1
        pbar.update(1)
    pbar.close()
    values = np.array(values)
    patches = np.array(patches)
    mean = np.mean(values,axis=0)
    std  = np.std(values,axis=0)
    p_mean = np.mean(patches,axis=0)
    #fig, ax = plt.subplots()
    #ax.scatter(p_mean,mean)
    #ax.set_xlabel('Percent of Total Patches')
    #ax.set_ylabel('Relative Error')
    #plt.show()
    return p_mean,mean,std

def test(s,bins=10,bin_range=(-1,0),reps=3,samples=100):
    pts = s.pv_polydata.points
    pts = [pt.tolist() for pt in pts]
    #pts = [pt.append(len(s.patches)) for pt in pts]
    values = []
    for pt in pts:
        data = pt
        data.append(len(s.patches))
        values.append(s.DD[0](data)[0])
    values = np.array(values).flatten()
    #s.depth_interior_min = min(values)
    s.depth_interior_max = max(values)
    bin_r = abs(bin_range[1] - bin_range[0])
    bin_range = [abs(bin_range[0])*s.depth_interior_min,abs(bin_range[0])*s.depth_interior_min+bin_r*abs(s.depth_interior_min - s.depth_interior_max)]
    BINS = np.linspace(bin_range[0],bin_range[1],bins+1)
    VALUES_MEAN = []
    VALUES_STD  = []
    PATCHES     = []
    for i in range(bins):
        p_mean,mean,std = get_values(s,BINS[i],BINS[i+1],reps,samples)
        VALUES_MEAN.append(mean)
        VALUES_STD.append(std)
        PATCHES.append(p_mean)
    #fig,ax = plt.subplots()
    #for i in range(bins):
    #    ax.scatter(PATCHES[i],VALUES_MEAN[i],label="{}-{}".format(BINS[i],BINS[i+1]))
    #ax.set_xlabel('Percent of Total Patches')
    #ax.set_ylabel('Relative Error')
    #plt.legend()
    #plt.show()
    return PATCHES,VALUES_MEAN,VALUES_STD

def convexity(surf_list):
    """
    Convexity measure obtained from paper below:

    CON  = V/V(CH)

    Shi, X., Li, R., Sheng, Y. (2020). A New Volume-Based Convexity Measure for
    3D Shapes. In: , et al. Advances in Computer Graphics. CGI 2020. Lecture
    Notes in Computer Science(), vol 12221. Springer, Cham.
    https://doi.org/10.1007/978-3-030-61864-3_6
    """
    con = []
    for surf in surf_list:
        delaunay = surf.pv_polydata.delaunay_3d()
        surf.convexity = surf.volume/delaunay.volume
        con.append(surf.convexity)
    return con

def gen_figure(surf_list,bins=1,bin_range=(-1,0),reps=10,samples=100):
    con = convexity(surf_list)
    c = np.array(con)
    norm = mpl.colors.Normalize(vmin=0.5, vmax=1)
    cmap = mpl.cm.ScalarMappable(norm=norm, cmap=mpl.cm.coolwarm)
    cmap.set_array([])
    surf_data = []
    fig,ax = plt.subplots()
    for surf in surf_list:
        p,vm,vs = test(surf,bins=bins,bin_range=bin_range,reps=reps,samples=samples)
        for i in range(bins):
            ax.scatter(p[i],vm[i],color=cmap.to_rgba(surf.convexity),marker=(2+i,2,0))
        surf_data.append((p,vm,vs))
    ax.set_xlabel('Percent of Total Patches')
    ax.set_ylabel('Relative Error')
    fig.colorbar(cmap,label=r"Convexity")
    name = "Patch_accuracy_convexity_bins{}_range{}_{}_reps{}.svg".format(bins,bin_range[0],bin_range[1],reps)
    fig.savefig(name,format="svg")
    plt.show()
    return surf_data
#gen_figure([cube,cylinder,heart,gyrus],bins=4)
"""
def test(s,bins=10,bin_range=(-1,0),reps=3):
    mesh = s.tet.grid
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    BINS = np.linspace(bin_range[0],bin_range[1],bins+1)
    VALUES = [[None]*reps]*bins
    done = False
    while not done:
        xb = mesh.bounds[1] - mesh.bounds[0]
        yb = mesh.bounds[3] - mesh.bounds[2]
        zb = mesh.bounds[5] - mesh.bounds[4]
        xp = (mesh.bounds[0] + np.random.random(1)*xb).item()
        yp = (mesh.bounds[2] + np.random.random(1)*yb).item()
        zp = (mesh.bounds[4] + np.random.random(1)*zb).item()
        tmp = s.DD[0]([xp,yp,zp,len(s.patches)])[0]
        tmp_ind = place_bin(tmp,bins,bin_range)
        if tmp_ind is None:
            continue
        if not None in VALUES[tmp_ind]:
            continue
        value  = []
        patch  = []
        for i in range(2,len(s.patches)):
            value.append(s.DD[0]([xp,yp,zp,i]))
            patch.append(i)
        value = np.array(value)
        name = value[-1].item()
        patch = np.array(patch)
        value = abs(value - value[-1])
        value = value/max(value)
        next_ind = VALUES[tmp_ind].index(None)
        VALUES[tmp_ind][next_ind] = value.tolist()
        flat_values = list(chain(*VALUES))
        if None not in flat_values:
            done = True
            break
        #ax.plot(patch,value,label=str(name))
    for i in range(bins):
        VALUE_bin = np.array(VALUES[i])
        VALUE_MEAN = np.mean(VALUE_bin,axis=0)
        VALUE_STD  = np.std(VALUE_bin,axis=0)
        ax.scatter(patch/patch[-1],VALUE_MEAN)
    #ax.fill_between(patch,VALUE_MEAN-VALUE_STD,VALUE_MEAN+VALUE_STD,alpha=0.5)
    plt.xlabel('Percent of Total Patches')
    plt.ylabel('Relative Error')
    plt.legend()
    plt.show()
    return VALUES
"""

#V = test(s)



import vtk
from vtk.util import numpy_support

def extract_surface(msh):
    msh_vtk = numpy_support.numpy_to_vtk(msh)
    pts = vtk.vtkPoints()
    pts.SetData(msh_vtk)
    poly = vtk.vtkPolyData()
    poly.SetPoints(pts)
    bounds = poly.GetBounds()
    x_range = bounds[1]-bounds[0]
    y_range = bounds[3]-bounds[2]
    z_range = bounds[5]-bounds[4]

    PCA = vtk.vtkPCANormalEstimation()
    PCA.SetInputData(poly)
    PCA.SetSampleSize(10)
    PCA.SetNormalOrientationToGraphTraversal()
    PCA.FlipNormalsOn()

    distance = vtk.vtkSignedDistance()
    distance.SetInputConnection(PCA.GetOutputPort())
    distance.SetRadius(1)
    #distance.SetDimensions(256,256,256)
    distance.SetBounds(bounds[0] - x_range * .1, bounds[1] + x_range * .1,
                       bounds[2] - y_range * .1, bounds[3] + y_range * .1,
                       bounds[4] - z_range * .1, bounds[5] + z_range * .1)
    surf_extract = vtk.vtkExtractSurface()
    surf_extract.SetInputConnection(distance.GetOutputPort())
    surf_extract.SetRadius(1)
    surf_extract.Update()
    clean = vtk.vtkCleanPolyData()
    clean.SetInputConnection(surf_extract.GetOutputPort())
    clean.Update()
    tri = vtk.vtkTriangleFilter()
    tri.SetInputConnection(clean.GetOutputPort())
    tri.Update()
    return tri.GetOutput()


def geodesic(polydata,proximal,distal,terminal,bifurcation):

    proximal_idx     = polydata.FindPoint(proximal)
    distal_idx       = polydata.FindPoint(distal)
    terminal_idx     = polydata.FindPoint(terminal)
    bifurcation_idx  = polydata.FindPoint(bifurcation)

    DIJ_1 = vtk.vtkDijkstraGraphGeodesicPath()
    DIJ_1.SetInputData(polydata)
    DIJ_1.SetStartVertex(proximal_idx)
    DIJ_1.SetEndVertex(bifurcation_idx)
    DIJ_1.Update()

    DIJ_2 = vtk.vtkDijkstraGraphGeodesicPath()
    DIJ_2.SetInputData(polydata)
    DIJ_2.SetStartVertex(bifurcation_idx)
    DIJ_2.SetEndVertex(distal_idx)
    DIJ_2.Update()

    DIJ_3 = vtk.vtkDijkstraGraphGeodesicPath()
    DIJ_3.SetInputData(polydata)
    DIJ_3.SetStartVertex(bifurcation_idx)
    DIJ_3.SetEndVertex(terminal_idx)
    DIJ_3.Update()


    poly_path_1 = DIJ_1.GetOutput()
    poly_path_2 = DIJ_2.GetOutput()
    poly_path_3 = DIJ_3.GetOutput()

    proximal_geodesic  = numpy_support.vtk_to_numpy(poly_path_1.GetPoints().GetData())
    distal_geodesic    = numpy_support.vtk_to_numpy(poly_path_2.GetPoints().GetData())
    terminal_geodesic  = numpy_support.vtk_to_numpy(poly_path_3.GetPoints().GetData())

    proximal_length = poly_path_1.GetLength()
    distal_length   = poly_path_2.GetLength()
    terminal_length = poly_path_3.GetLength()

    proximal_data = {'path'  :proximal_geodesic,
                     'length':proximal_length}

    distal_data   = {'path'  :distal_geodesic,
                     'length':distal_length}

    terminal_data = {'path'  :terminal_geodesic,
                     'length':terminal_length}

    return proximal_data,distal_data,terminal_data



import numpy as np
import numba as nb

@nb.jit(nopython=True,fastmath=True)
def fitness(x,H_0,dim=3):
    n = H_0.shape[0]//dim
    a = np.ones(n*dim)
    for i in range(n):
        a[i] = np.cos(x[i*2+1])*np.sin(x[i*2])
        a[i+n] = np.sin(x[i*2+1])*np.sin(x[i*2])
        a[i+2*n] = np.cos(x[i*2])
    return [a.T@H_0@a]



import numba as nb

@nb.jit(nopython=True,cache=True,nogil=True)
def add_depths(data, edge):
    search = [int(edge)]
    while len(search) > 0:
        for idx in search:
            if data[idx, 15].item() >= 0 or data[idx, 16].item() >= 0:
                left = int(data[idx, 15].item())
                right = int(data[idx, 16].item())
                data[left, 26] += 1
                data[right, 26] += 1
                search.remove(idx)
                search.append(left)
                search.append(right)
            else:
                search.remove(idx)


"""
Code to evaluate cache hits and misses for svcco
"""
import svcco
import numpy as np
from binarytree import Node
from time import time, perf_counter
from functools import lru_cache
import matplotlib.pyplot as plt
import os
import cProfile
import pstats
from memory_profiler import memory_usage
from copy import deepcopy

os.environ["PATH"] += ';C:\\Program Files (x86)\\Graphviz\\bin'
os.environ["PATH"] += ';C:\\Program Files (x86)\\Graphviz'
################################################
# CODE FOR LINKED LIST CONTAINER
################################################

def Vessel(data):
    node                = Node(int(data[-1]))
    node.data           = np.array(deepcopy(data.reshape(1,-1).tolist()))
    node.proximal_point = data[0:3]
    node.distal_point   = data[3:6]
    node.radius         = data[21]
    node.length         = data[20]
    node.left_bif       = data[23]
    node.right_bif      = data[24]
    node.daughter_0     = int(data[15])
    node.daughter_1     = int(data[16])
    node.parent         = int(data[17])
    node.id             = int(data[-1])
    return node

class Tree:
    # Constructor
    def __init__(self):
        self.root = None
        self.all_nodes = []
        self.all_data  = []
        self.all_data_list = []
    def insert(self,data):
        new_vessel = Vessel(data)
        if new_vessel.id == 0:
            new_vessel.parent_node = None
            self.root = new_vessel
            self.all_nodes.append(new_vessel)
            self.all_data.append(np.array(deepcopy(new_vessel.data[0,:].tolist())))
            self.all_data_list.append(new_vessel.data[0,:].tolist())
        else:
            queue = [self.root]
            while queue:
                vessel = queue.pop(0)
                if vessel.id == new_vessel.parent:
                    if vessel.daughter_0 == new_vessel.id:
                        new_vessel.parent_node = vessel
                        vessel.left = new_vessel
                        self.all_nodes.append(new_vessel)
                        self.all_data.append(np.array(deepcopy(new_vessel.data[0,:].tolist())))
                        self.all_data_list.append(new_vessel.data[0,:].tolist())
                        break
                    if vessel.daughter_1 == new_vessel.id:
                        new_vessel.parent_node = vessel
                        vessel.right = new_vessel
                        self.all_nodes.append(new_vessel)
                        self.all_data.append(np.array(deepcopy(new_vessel.data[0,:].tolist())))
                        self.all_data_list.append(new_vessel.data[0,:].tolist())
                        break
                if vessel.daughter_0 > 0:
                    queue.append(vessel.left)
                if vessel.daughter_1 > 0:
                    queue.append(vessel.right)

    def find_closest(self,point,method='fbs'):
        best_value = np.inf
        best_id    = None
        if method == 'fbs':
            queue = [self.root]
            while queue:
                vessel = queue.pop(0)
                _,dist = svcco.branch_addition.close.close_exact(vessel.data,point)
                if dist < best_value:
                    best_value = dist
                    best_id    = vessel.id
                if vessel.daughter_0 > 0:
                    queue.append(vessel.left)
                if vessel.daughter_1 > 0:
                    queue.append(vessel.right)
        elif method == 'all':
            for vessel in self.all_nodes:
                _,dist = svcco.branch_addition.close.close_exact(vessel.data,point)
                if dist < best_value:
                    best_value = dist
                    best_id    = vessel.id
        elif method == 'vec':
            #data = [v.data[0,:] for v in self.root.preorder]
            best_id,best_value = svcco.branch_addition.close.close_binary_vectorize(self.all_data,point)
            best_id = int(best_id)
        return best_id,best_value

    def update_radii(self,factor):
        self.root.radius *= factor
        queue = [self.root]
        while queue:
            vessel = queue.pop(0)
            if vessel.daughter_0 > 0:
                vessel.left.radius = vessel.radius*vessel.left_bif
                queue.append(vessel.left)
            elif vessel.daughter_1 > 0:
                vessel.right.radius = vessel.radius*vessel.right_bif
                queue.append(vessel.right)

    def find_collision(self,new_vessel,method='fbs'):
        if method == 'fbs':
            collisions = False
            queue = [self.root]
            while queue:
                vessel = queue.pop(0)
                if len( svcco.collision.sphere_proximity.sphere_proximity_testing(vessel.data,new_vessel)) > 0:
                    collisions = svcco.collision.obb.obb(vessel.data,new_vessel)
                    if collisions:
                        return True
                if vessel.daughter_0 > 0:
                    queue.append(vessel.left)
                if vessel.daughter_1 > 0:
                    queue.append(vessel.right)
        elif method == 'all':
            collisions = []
            for vessel in self.all_nodes:
                if len(svcco.collision.sphere_proximity.sphere_proximity_testing(vessel.data,new_vessel)) > 0:
                    collisions.append(svcco.collision.obb.obb(vessel.data,new_vessel))
            return any(collisions)
        #elif method == 'vec':
        #    #data = [v.data[0,:] for v in self.root.preorder]
        #    best_id,best_value = svcco.branch_addition.close.close_binary_vectorize(self.all_data,point)
        #    best_id = int(best_id)
        #return collision

def build(data):
    t = Tree()
    t.insert(data[0,:])
    queue = []
    if data[0,15] > 0:
        queue.append(int(data[0,15]))
    if data[0,16] > 0:
        queue.append(int(data[0,16]))
    while queue:
        id = queue.pop(0)
        t.insert(data[id,:])
        if data[id,15] > 0:
            queue.append(int(data[id,15]))
        if data[id,16] > 0:
            queue.append(int(data[id,16]))
    return t
#####################################################
# CODE FOR BUILDING SVCCO Tree
#####################################################
import pyvista as pv

cube = pv.Cube().triangulate().subdivide(3)

s = svcco.surface()
s.set_data(10*cube.points,normals=cube.point_normals)
s.solve()
s.build()

t = svcco.tree()
t.set_boundary(s)
t.set_root()
t.convex = True
#####################################################
# WRAPPERS FOR TESTING CALLS
#####################################################

# FIND CLOSEST VESSEL
def test_sv_tree(sv_tree,point):
    start = perf_counter()
    best_id,best_dist = svcco.branch_addition.close.close_exact(sv_tree.data,point)
    elapsed = perf_counter() - start
    best_id = best_id[0]
    return elapsed,best_id

"""
def test_sv_tree_calls(t,point):
    cProfile.run("svcco.branch_addition.close.close_exact(t.data,point)","sv_stats")
    p = pstats.Stats("sv_stats")
    return p.total_calls
"""

def test_sv_tree_mem(t):
    mem = t.data.nbytes
    return mem*10e-9


def test_binary_tree(binary,point,meth='fbs'):
    start = perf_counter()
    best_id,best_dist = binary.find_closest(point,method=meth)
    elapsed = perf_counter() - start
    return elapsed,best_id

def test_binary_mem(binary):
    locs = []
    for vessel in binary.root.preorder:
        #locs.append(vessel.data.__array_interface__['data'][0])
        locs.append(vessel.data.ctypes.data)
    mem = max(locs)-min(locs)
    if mem == 0:
        mem = binary.root.data.nbytes
    return mem*10e-9
"""
def test_binary_calls(binary,point,meth='fbs'):
    cProfile.run("binary.find_closest(point,method='{}')".format(meth),"bin_stats")
    p = pstats.Stats("bin_stats")
    return p.total_calls

def test_binary_mem(binary,point,meth='fbs'):
    mem = memory_usage((binary.find_closest,(point,),{'method':meth}))
    return np.mean(mem)
"""

# UPDATE TREE RADII (rescale)

def test_sv_radii_time(t,factor):
    start = perf_counter()
    t.data[:,21] = t.data[0,21]*factor*t.data[:,28]
    elapsed = perf_counter()-start
    return elapsed

def test_binary_radii_time(binary,factor):
    start = perf_counter()
    binary.update_radii(factor)
    elapsed = perf_counter()-start
    return elapsed

# TEST TREE COLLISIONS
def test_sv_collision_time(t,vessel,meth='fbs'):
    if meth == 'fbs':
        start = perf_counter()
        n = svcco.collision.sphere_proximity.sphere_proximity_testing(t.data,vessel)
        collision = svcco.collision.obb.obb(t.data[n,:],vessel)
        elapsed = perf_counter()-start
    else:
        start = perf_counter()
        n = svcco.collision.sphere_proximity.sphere_proximity_testing(t.data,vessel)
        collision = svcco.collision.obb.obbc(t.data[n,:],vessel)
        elapsed = perf_counter()-start
    return elapsed

def test_binary_collision_time(binary,vessel,meth='fbs'):
    start = perf_counter()
    collision = binary.find_collision(vessel,method=meth)
    elapsed = perf_counter() - start
    return elapsed


#####################################################
# CODE FOR TESTING PERFORMANCE
#####################################################

binary = build(t.data)
point = np.array([0.2,0.2,0.2])

def test(t,binary,point,reps=20):
    sv_tree_perf = []
    binary_tree_perf_fbs = []
    binary_tree_perf_all = []
    binary_tree_perf_vec = []
    #sv_mem = test_sv_tree_mem(t,point)
    #bin_mem = test_binary_mem(binary,point,meth='vec')
    elapsed_binary_vec,best_binary = test_binary_tree(binary,point,meth='vec')
    for i in range(reps):
        point = np.random.random(3)*10-5
        elapsed_sv,best_sv = test_sv_tree(t,point)
        sv_tree_perf.append(elapsed_sv)
        elapsed_binary_fbs,best_binary = test_binary_tree(binary,point)
        binary_tree_perf_fbs.append(elapsed_binary_fbs)
        elapsed_binary_all,best_binary = test_binary_tree(binary,point,meth='all')
        binary_tree_perf_all.append(elapsed_binary_all)
        elapsed_binary_vec,best_binary = test_binary_tree(binary,point,meth='vec')
        binary_tree_perf_vec.append(elapsed_binary_vec)
        if best_sv != best_binary:
            print('ERROR: ANSWERS DO NOT MATCH {} != {}'.format(best_sv,best_binary))
            break
    #sv_mem = test_sv_tree_mem(t,point)
    #bin_mem = test_binary_mem(binary,point,meth='vec')
    #sv_calls = test_sv_tree_calls(t,point)
    #bin_calls = test_binary_calls(binary,point,meth='vec')
    print('PERFORMANCE, REPITITIONS = {}'.format(reps))
    print('-------------------------------------------------------')
    print('SV TREE PERFORMANCE             : {}'.format(np.mean(sv_tree_perf)))
    print('BINARY TREE PERFORMANCE METH=FBS: {}'.format(np.mean(binary_tree_perf_fbs)))
    print('BINARY TREE PERFORMANCE METH=ALL: {}'.format(np.mean(binary_tree_perf_all)))
    print('BINARY TREE PERFORMANCE METH=VEC: {}'.format(np.mean(binary_tree_perf_vec)))
    print('-------------------------------------------------------')
    #print('INSTRUCTION PERFORMANCE')
    #print('-------------------------------------------------------')
    #print('SV TREE PERFORMANCE    : {}'.format(sv_calls))
    #print('BINARY TREE PERFORMANCE: {}'.format(bin_calls))
    #print('-------------------------------------------------------')
    #print('MEMORY PERFORMANCE')
    #print('-------------------------------------------------------')
    #print('SV TREE PERFORMANCE    : {}'.format(sv_mem))
    #print('BINARY TREE PERFORMANCE: {}'.format(bin_mem))
    #print('-------------------------------------------------------')
    return np.mean(sv_tree_perf),np.mean(binary_tree_perf_fbs),np.mean(binary_tree_perf_all),np.mean(binary_tree_perf_vec) #,sv_calls,bin_calls,sv_mem,bin_mem

def test_radii(t,binary,reps=20):
    sv_time = []
    binary_time = []
    for rep in range(reps):
        factor = np.random.random(1)[0]
        sv_time.append(test_sv_radii_time(t,factor))
        binary_time.append(test_binary_radii_time(binary,factor))
    print('PERFORMANCE, REPITITIONS = {}'.format(reps))
    print('-------------------------------------------------------')
    print('SV TREE PERFORMANCE    : {}'.format(np.mean(sv_time)))
    print('BINARY TREE PERFORMANCE: {}'.format(np.mean(binary_time)))
    print('-------------------------------------------------------')
    return np.mean(sv_time), np.mean(binary_time)

def test_collision(t,binary,reps=20):
    sv_time = []
    sv_all = []
    binary_time_fbs = []
    binary_time_all = []
    for rep in range(reps):
        P0 = np.random.random(3)*10-5
        P1 = np.random.random(3)*10-5
        L  = np.linalg.norm(P0-P1)
        R  = np.random.random(1)*2
        vessel = np.ones(30)*-1
        vessel[0:3] = P0
        vessel[3:6] = P1
        vessel[20]  = L
        vessel[21]  = R
        test_sv_collision_time(t,vessel)
        test_sv_collision_time(t,vessel,meth='all')
        test_binary_collision_time(binary,vessel)
        test_binary_collision_time(binary,vessel,meth='all')
        sv_time.append(test_sv_collision_time(t,vessel))
        sv_all.append(test_sv_collision_time(t,vessel,meth='all'))
        binary_time_fbs.append(test_binary_collision_time(binary,vessel))
        binary_time_all.append(test_binary_collision_time(binary,vessel,meth='all'))
    print('PERFORMANCE, REPITITIONS = {}'.format(reps))
    print('-------------------------------------------------------')
    print('SV TREE PERFORMANCE         : {}'.format(np.mean(sv_time)))
    print('SV TREE PERFORMANCE ALL     : {}'.format(np.mean(sv_all)))
    print('BINARY TREE PERFORMANCE  FBS: {}'.format(np.mean(binary_time_fbs)))
    print('BINARY TREE PERFORMANCE  ALL: {}'.format(np.mean(binary_time_all)))
    print('-------------------------------------------------------')
    return np.mean(sv_time), np.mean(sv_all), np.mean(binary_time_fbs), np.mean(binary_time_all)

def test_range(start=1,stop=1000,incr=10,show=False):
    t = svcco.tree()
    t.set_boundary(s)
    t.set_root()
    t.convex = True
    binary = build(t.data)
    SIZE     = [start]
    SV       = []
    BIN_FBS  = []
    BIN_ALL  = []
    BIN_VEC  = []
    #SV_CALLS = []
    #BIN_CALLS= []
    #SV_MEM   = []
    #BIN_MEM  = []
    while SIZE[-1] < stop:
        sv_perf,bin_perf_fbs,bin_perf_all,bin_perf_vec = test(t,binary,point)
        SV.append(sv_perf)
        BIN_FBS.append(bin_perf_fbs)
        BIN_ALL.append(bin_perf_all)
        BIN_VEC.append(bin_perf_vec)
        #SV_CALLS.append(sv_calls)
        #BIN_CALLS.append(bin_calls)
        #SV_MEM.append(sv_mem)
        #BIN_MEM.append(bin_mem)
        t.n_add(incr)
        binary = build(t.data)
        SIZE.append(SIZE[-1]+incr)
    SIZE.pop(-1)
    if show:
        fig = plt.figure()
        ax  = fig.add_subplot(111)
        ax.plot(SIZE,SV,label="SV continuous")
        ax.plot(SIZE,BIN_FBS,label="Binary Tree FBS")
        #ax.plot(SIZE,BIN_ALL,label="Binary Tree Precollected")
        ax.plot(SIZE,BIN_VEC,label="Binary Tree Vectorized FBS")
        ax.set_xlabel('Tree Size')
        ax.set_ylabel('Time (seconds)')
        plt.legend()
        plt.show()
    return SIZE,SV,BIN_FBS,BIN_ALL,BIN_VEC

def test_range_radii(start=1,stop=1000,incr=10,show=False):
    t = svcco.tree()
    t.set_boundary(s)
    t.set_root()
    t.convex = True
    binary = build(t.data)
    SIZE = [start]
    SV = []
    BIN = []
    while SIZE[-1] < stop:
        sv_t,bin_t = test_radii(t,binary)
        SV.append(sv_t)
        BIN.append(bin_t)
        t.n_add(incr)
        SIZE.append(SIZE[-1]+incr)
        binary = build(t.data)
    SIZE.pop(-1)
    if show:
        fig = plt.figure()
        ax  = fig.add_subplot(111)
        ax.plot(SIZE,SV,label="SV")
        ax.plot(SIZE,BIN,label="Binary Tree")
        ax.set_xlabel("Tree Size")
        ax.set_ylabel("Time (seconds)")
        plt.legend()
        plt.show()
    return SIZE,SV,BIN

def test_range_collision(start=1,stop=1000,incr=10,show=False):
    t = svcco.tree()
    t.set_boundary(s)
    t.set_root()
    t.convex = True
    binary = build(t.data)
    SIZE = [start]
    SV = []
    SV_ALL = []
    BIN_FBS = []
    BIN_ALL = []
    SV_MEM  = []
    BIN_MEM = []
    while SIZE[-1] < stop:
        sv_t,sv_all,bin_fbs,bin_all = test_collision(t,binary)
        SV.append(sv_t)
        SV_ALL.append(sv_all)
        BIN_FBS.append(bin_fbs)
        BIN_ALL.append(bin_all)
        SV_MEM.append(test_sv_tree_mem(t))
        BIN_MEM.append(test_binary_mem(binary))
        t.n_add(incr)
        SIZE.append(SIZE[-1]+incr)
        binary = build(t.data)
    SIZE.pop(-1)
    if show:
        fig = plt.figure()
        ax  = fig.add_subplot(121)
        ax.plot(SIZE,SV,label="SV")
        ax.plot(SIZE,SV_ALL,label="SV all")
        ax.plot(SIZE,BIN_FBS,label="Binary Tree FBS")
        ax.plot(SIZE,BIN_ALL,label="Binary Tree ALL")
        ax.set_xlabel("Tree Size")
        ax.set_ylabel("Time (seconds)")
        ax1  = fig.add_subplot(122)
        ax1.plot(SIZE,SV_MEM,label="SV Memory")
        ax1.plot(SIZE,BIN_MEM,label="Binary Memory")
        ax1.set_xlabel("Tree Size")
        ax1.set_ylabel("Memory Span Locality(Gb)")
        plt.legend()
        plt.show()
    return SIZE,SV,SV_ALL,BIN_FBS,BIN_ALL,SV_MEM,BIN_MEM

def test_meta(num,stop=1000,incr=10):

    size,sv,bin_fbs,bin_all,bin_vec = test_range(stop=stop,incr=incr)
    SIZE = np.array(size)
    SV   = np.array(sv)
    BIN_FBS = np.array(bin_fbs)
    BIN_ALL = np.array(bin_all)
    BIN_VEC = np.array(bin_vec)
    for i in range(1,num):
        size,sv,bin_fbs,bin_all,bin_vec = test_range(stop=stop,incr=incr)
        SV      = np.vstack((SV,np.array(sv)))
        BIN_FBS = np.vstack((BIN_FBS,np.array(bin_fbs)))
        BIN_ALL = np.vstack((BIN_ALL,np.array(bin_all)))
        BIN_VEC = np.vstack((BIN_VEC,np.array(bin_vec)))
    SV_MEAN      = np.mean(SV,axis=0)
    BIN_FBS_MEAN = np.mean(BIN_FBS,axis=0)
    BIN_ALL_MEAN = np.mean(BIN_ALL,axis=0)
    BIN_VEC_MEAN = np.mean(BIN_VEC,axis=0)
    SV_STD       = np.std(SV,axis=0)
    BIN_FBS_STD  = np.std(BIN_FBS,axis=0)
    BIN_ALL_STD  = np.std(BIN_ALL,axis=0)
    BIN_VEC_STD  = np.std(BIN_VEC,axis=0)
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    ax.plot(SIZE,SV_MEAN,label="SV")
    ax.fill_between(SIZE,SV_MEAN-SV_STD,SV_MEAN+SV_STD,alpha=0.5)
    ax.plot(SIZE,BIN_FBS_MEAN,label="Binary Tree FBS")
    ax.fill_between(SIZE,BIN_FBS_MEAN-BIN_FBS_STD,BIN_FBS_MEAN+BIN_FBS_STD,alpha=0.5)
    ax.plot(SIZE,BIN_VEC_MEAN,label="Binary Tree Vectorized FBS")
    ax.fill_between(SIZE,BIN_VEC_MEAN-BIN_VEC_STD,BIN_VEC_MEAN+BIN_VEC_STD,alpha=0.5)
    ax.set_xlabel('Tree Size')
    ax.set_ylabel('Time (seconds)')
    plt.legend(loc='upper left')
    plt.savefig('C:\\Users\\Zack\\Downloads\\FIGURE 1 PART C.png')

    size,sv,bin = test_range_radii(stop=stop,incr=incr)
    SIZE = np.array(size)
    SV   = np.array(sv)
    BIN = np.array(bin)
    for i in range(1,num):
        size,sv,bin = test_range_radii(stop=stop,incr=incr)
        SV      = np.vstack((SV,np.array(sv)))
        BIN     = np.vstack((BIN,np.array(bin)))
    SV_MEAN      = np.mean(SV,axis=0)
    BIN_MEAN     = np.mean(BIN,axis=0)
    SV_STD       = np.std(SV,axis=0)
    BIN_STD      = np.std(BIN,axis=0)
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    ax.plot(SIZE,SV_MEAN,label="SV")
    ax.fill_between(SIZE,SV_MEAN-SV_STD,SV_MEAN+SV_STD,alpha=0.5)
    ax.plot(SIZE,BIN_MEAN,label="Binary Tree")
    ax.fill_between(SIZE,BIN_MEAN-BIN_STD,BIN_MEAN+BIN_STD,alpha=0.5)
    ax.set_xlabel('Tree Size')
    ax.set_ylabel('Time (seconds)')
    plt.legend(loc='upper left')
    plt.savefig('C:\\Users\\Zack\\Downloads\\FIGURE 1 PART D.png')

    size,sv,sv_all,bin_fbs,bin_all,sv_mem,bin_mem = test_range_collision(stop=stop,incr=incr)
    SIZE = np.array(size)
    SV   = np.array(sv)
    SV_ALL = np.array(sv_all)
    BIN_FBS = np.array(bin_fbs)
    BIN_ALL = np.array(bin_all)
    SV_MEM  = np.array(sv_mem)
    BIN_MEM = np.array(bin_mem)
    for i in range(1,num):
        size,sv,sv_all,bin_fbs,bin_all,sv_mem,bin_mem = test_range_collision(stop=stop,incr=incr)
        SV      = np.vstack((SV,np.array(sv)))
        SV_ALL  = np.vstack((SV_ALL,np.array(sv)))
        BIN_FBS = np.vstack((BIN_FBS,np.array(bin_fbs)))
        BIN_ALL = np.vstack((BIN_ALL,np.array(bin_all)))
        SV_MEM  = np.vstack((SV_MEM,np.array(sv_mem)))
        BIN_MEM = np.vstack((BIN_MEM,np.array(bin_mem)))
    SV_MEAN      = np.mean(SV,axis=0)
    SV_ALL_MEAN  = np.mean(SV_ALL,axis=0)
    BIN_FBS_MEAN = np.mean(BIN_FBS,axis=0)
    BIN_ALL_MEAN = np.mean(BIN_ALL,axis=0)
    SV_MEM_MEAN  = np.mean(SV_MEM,axis=0)
    BIN_MEM_MEAN = np.mean(BIN_MEM,axis=0)
    SV_STD       = np.std(SV,axis=0)
    SV_ALL_STD   = np.std(SV_ALL,axis=0)
    BIN_FBS_STD  = np.std(BIN_FBS,axis=0)
    BIN_ALL_STD  = np.std(BIN_ALL,axis=0)
    SV_MEM_STD   = np.std(SV_MEM,axis=0)
    BIN_MEM_STD  = np.std(BIN_MEM,axis=0)
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    ax.plot(SIZE,SV_MEAN,label="SV FBS")
    ax.fill_between(SIZE,SV_MEAN-SV_STD,SV_MEAN+SV_STD,alpha=0.5)
    ax.plot(SIZE,SV_ALL_MEAN,label="SV all")
    ax.fill_between(SIZE,SV_ALL_MEAN-SV_ALL_STD,SV_ALL_MEAN+SV_ALL_STD,alpha=0.5)
    ax.plot(SIZE,BIN_FBS_MEAN,label="Binary Tree FBS")
    ax.fill_between(SIZE,BIN_FBS_MEAN-BIN_FBS_STD,BIN_FBS_MEAN+BIN_FBS_STD,alpha=0.5)
    ax.plot(SIZE,BIN_ALL_MEAN,label="Binary Tree ALL")
    ax.fill_between(SIZE,BIN_ALL_MEAN-BIN_ALL_STD,BIN_ALL_MEAN+BIN_ALL_STD,alpha=0.5)
    ax.set_xlabel('Tree Size')
    ax.set_ylabel('Time (seconds)')
    ax.legend(loc='upper left')
    ax.set_ylim(bottom=0)
    plt.savefig('C:\\Users\\Zack\\Downloads\\FIGURE 1 PART E.png')

    fig = plt.figure()
    ax  = fig.add_subplot(111)
    ax.plot(SIZE,SV_MEM_MEAN,label="SV Memory")
    ax.fill_between(SIZE,SV_MEM_MEAN-SV_MEM_STD,SV_MEM_MEAN+SV_MEM_STD,alpha=0.5)
    ax.plot(SIZE,BIN_MEM_MEAN,label="Binary Memory")
    ax.fill_between(SIZE,BIN_MEM_MEAN-BIN_MEM_STD,BIN_MEM_MEAN+BIN_MEM_STD,alpha=0.5)
    ax.legend(loc='upper left')
    ax.set_xlabel('Tree Size')
    ax.set_ylabel('Span of Memory Locality (Gbs)')
    ax.set_ylim(bottom=0)
    plt.savefig('C:\\Users\\Zack\\Downloads\\FIGURE 1 PART F.png')

"""
import networkx
import pydotplus

def to_networkx(graph):
    dotplus =  pydotplus.graph_from_dot_data(graph.source)
    nx_graph = networkx.nx_pydot.from_pydot(dotplus)
    return nx_graph
"""



import numpy as np
from .update import update,naive
from .calculate_radii import radii
from .calculate_length import length
from .add_depths import add_depths
from .add_flow import add_flow
from .basis import basis
import vtk

def finite_difference(data,precomputed,terminal,vessel,gamma,nu,Qterm,Pperm,Pterm):
    volumes =  []
    trials = []
    #print('parents initial: {}'.format(data[:,17]))
    for i in range(precomputed.shape[0]):
        trial_data = np.vstack((data,np.zeros((2,data.shape[1]))))
        trial_data[-1,0:3] = precomputed[i,:]
        trial_data[-2,0:3] = precomputed[i,:]
        trial_data[-2,3:6] = terminal
        trial_data[-1,3:6] = trial_data[vessel,3:6]
        trial_data[vessel,3:6] = precomputed[i,:]
        trial_data[-2,15] = -1.0
        trial_data[-2,16] = -1.0
        trial_data[-2,18] = max(trial_data[:,19]) + 1
        trial_data[-2,19] = max(trial_data[:,19]) + 2
        trial_data[-1,18] = trial_data[-2,18]
        trial_data[-1,19] = trial_data[vessel,19]
        trial_data[-1,15] = trial_data[vessel,15]
        trial_data[-1,16] = trial_data[vessel,16]
        trial_data[-2,17] = vessel
        trial_data[-1,17] = vessel
        trial_data[vessel,15] = trial_data.shape[0]-2
        trial_data[vessel,16] = trial_data.shape[0]-1
        trial_data[vessel,19] = trial_data[-2,18]
        trial_data[-2,22] = Qterm
        trial_data[-1,22] = trial_data[vessel,22]
        trial_data[-2,-1] = trial_data.shape[0] - 2
        trial_data[-1,-1] = trial_data.shape[0] - 1
        trial_data[-2,26] = trial_data[vessel,26]
        trial_data[-1,26] = trial_data[vessel,26]
        if trial_data[-1,15] >= 0:
            child = int(trial_data[-1,15])
            trial_data[child,17] = trial_data[-1,-1]
        if trial_data[-1,16] >= 0:
            child = int(trial_data[-1,16])
            trial_data[child,17] = trial_data[-1,-1]
        length(trial_data,-1)
        length(trial_data,-2)
        length(trial_data,vessel)
        basis(trial_data,-1)
        basis(trial_data,-2)
        basis(trial_data,vessel)
        add_depths(trial_data,vessel)
        updated_flows = add_flow(trial_data,vessel,Qterm)
        #print('Flow')
        #print(trial_data[:,22])
        #print('Res')
        #print(trial_data[:,25])
        #print('lengths')
        #print(trial_data[:,20])
        #print('right daughters')
        #print(trial_data[:,15])
        #print('left daughters')
        #print(trial_data[:,16])
        #print('depths')
        #print(trial_data[:,26])
        update(trial_data,gamma,nu)
        updated_radii = radii(trial_data,Pperm,Pterm)
        # calculate tree volume
        volumes.append(np.sum(np.pi*trial_data[:,21]**2*trial_data[:,20]))
        trials.append(trial_data)
    idx = np.argmin(volumes)
    flow_check = []
    #parent = vessel
    #while parent >= 0:
    #    #data[parent,22] = flow + data[parent,22]
    #    flow_check.append(parent)
    #    parent = int(data[parent,17].item())
    #print(flow_check)
    fd_data = trials[idx]
    #print('parents: {}'.format(fd_data[:,17]))
    """
    colors = vtk.vtkNamedColors()
    models = []
    actors = []
    background_color='white'
    resolution=100
    #print(updated_flows)
    for edge in range(fd_data.shape[0]):
        center = tuple((fd_data[edge,0:3] + fd_data[edge,3:6])/2)
        radius = fd_data[edge,21]
        direction = tuple(fd_data[edge,12:15])
        vessel_length = fd_data[edge,20]
        cyl = vtk.vtkTubeFilter()
        line = vtk.vtkLineSource()
        line.SetPoint1(fd_data[edge,0],fd_data[edge,1],fd_data[edge,2])
        line.SetPoint2(fd_data[edge,3],fd_data[edge,4],fd_data[edge,5])
        cyl.SetInputConnection(line.GetOutputPort())
        cyl.SetRadius(radius)
        cyl.SetNumberOfSides(resolution)
        models.append(cyl)
        mapper = vtk.vtkPolyDataMapper()
        actor  = vtk.vtkActor()
        mapper.SetInputConnection(cyl.GetOutputPort())
        actor.SetMapper(mapper)
        if edge in updated_flows and edge != vessel:
            actor.GetProperty().SetColor(colors.GetColor3d('blue'))
        elif edge == vessel:
            actor.GetProperty().SetColor(colors.GetColor3d('green'))
        else:
            actor.GetProperty().SetColor(colors.GetColor3d('red'))
        #if edge in updated_radii:
        #    actor.GetProperty().SetColor(colors.GetColor3d('yellow'))
        actors.append(actor)
    renderer = vtk.vtkRenderer()
    renderer.SetBackground(colors.GetColor3d(background_color))

    render_window = vtk.vtkRenderWindow()
    render_window.AddRenderer(renderer)
    render_window.SetWindowName('SimVascular Vessel Flow Updating')

    interactor = vtk.vtkRenderWindowInteractor()
    interactor.SetRenderWindow(render_window)
    for actor in actors:
        renderer.AddActor(actor)

    render_window.Render()
    interactor.Start()
    """
    return precomputed[idx,:],idx,volumes,trials[idx]

def finite_difference_naive(data,precomputed,terminal,vessel,gamma,nu,Qterm,Pperm,Pterm):
    volumes =  []
    trials = []
    #print('parents initial: {}'.format(data[:,17]))
    for i in range(precomputed.shape[0]):
        trial_data = np.vstack((data,np.zeros((2,data.shape[1]))))
        trial_data[-1,0:3] = precomputed[i,:]
        trial_data[-2,0:3] = precomputed[i,:]
        trial_data[-2,3:6] = terminal
        trial_data[-1,3:6] = trial_data[vessel,3:6]
        trial_data[vessel,3:6] = precomputed[i,:]
        trial_data[-2,15] = -1.0
        trial_data[-2,16] = -1.0
        trial_data[-2,18] = max(trial_data[:,19]) + 1
        trial_data[-2,19] = max(trial_data[:,19]) + 2
        trial_data[-1,18] = trial_data[-2,18]
        trial_data[-1,19] = trial_data[vessel,19]
        trial_data[-1,15] = trial_data[vessel,15]
        trial_data[-1,16] = trial_data[vessel,16]
        trial_data[-2,17] = vessel
        trial_data[-1,17] = vessel
        trial_data[vessel,15] = trial_data.shape[0]-2
        trial_data[vessel,16] = trial_data.shape[0]-1
        trial_data[vessel,19] = trial_data[-2,18]
        trial_data[-2,22] = Qterm
        trial_data[-1,22] = trial_data[vessel,22]
        trial_data[-2,-1] = trial_data.shape[0] - 2
        trial_data[-1,-1] = trial_data.shape[0] - 1
        trial_data[-2,26] = trial_data[vessel,26]
        trial_data[-1,26] = trial_data[vessel,26]
        if trial_data[-1,15] >= 0:
            child = int(trial_data[-1,15])
            trial_data[child,17] = trial_data[-1,-1]
        if trial_data[-1,16] >= 0:
            child = int(trial_data[-1,16])
            trial_data[child,17] = trial_data[-1,-1]
        length(trial_data,-1)
        length(trial_data,-2)
        length(trial_data,vessel)
        basis(trial_data,-1)
        basis(trial_data,-2)
        basis(trial_data,vessel)
        add_depths(trial_data,vessel)
        updated_flows = add_flow(trial_data,vessel,Qterm)
        naive(trial_data,gamma,nu,0)
        updated_radii = radii(trial_data,Pperm,Pterm)
        # calculate tree volume
        volumes.append(np.sum(np.pi*trial_data[:,21]**2*trial_data[:,20]))
        trials.append(trial_data)
    idx = np.argmin(volumes)
    flow_check = []
    fd_data = trials[idx]

    return precomputed[idx,:],idx,volumes,trials[idx]



import svcco
import numpy as np
import matplotlib.pyplot as plt
import pyvista as pv
from tqdm import tqdm

cube = pv.Cube(x_length=3.72,y_length=3.72,z_length=3.72).triangulate().subdivide(3)

s = svcco.surface()
s.set_data(cube.points,cube.point_normals)
s.solve()
s.build(q=2) #,resolution=30)
res_cube = svcco.implicit.visualize.visualize.plot_mpu_boundaries(s,name="cube_point_enclosure",resolution=80,test_size=10000)

heart = svcco.surface()
heart_points = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_points_unique.csv',delimiter=',')
heart_normals = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_normals_unique.csv',delimiter=',')
heart.set_data(heart_points,heart_normals)
heart.solve()
heart.build(q=6) #,resolution=85,k=200,buffer=5)
res_heart = svcco.implicit.visualize.visualize.plot_mpu_boundaries(heart,name="heart_point_enclosure",resolution=80,plane_axis=0,test_size=10000)

disk = pv.Disc(r_res=20,c_res=100)
cyl  = disk.extrude([0,0,1],capping=True).triangulate()
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl)
cyl  = cyl.subdivide(2)
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl,hausd=0.005)
cyl  = cyl.compute_normals(auto_orient_normals=True,feature_angle=90)
c = svcco.surface()
c.set_data(cyl.points,cyl.point_normals)
c.solve()
c.build(q=8) #,k=200,resolution=100)
res_cyl = svcco.implicit.visualize.visualize.plot_mpu_boundaries(c,name="cylinder_point_enclosure",resolution=80,test_size=10000)

left_gyrus   = "D:\\Tree\\Tree_8-0\\brain_testing\\FJ3801_BP58201_FMA72658_Left inferior frontal gyrus.obj"

gyrus = svcco.surface()
gyrus.load(left_gyrus)
gyrus.solve()
gyrus.build(q=4) #,resolution=40,buffer=5)
res_gyrus = svcco.implicit.visualize.visualize.plot_mpu_boundaries(gyrus,name="gyrus_point_enclosure",resolution=80,test_size=10000)



import os
import sys
import platform
import glob

class file_options:

    def __init__(self,num_caps,time=[0,1],flow=[1,1],gui=True,distal_resistance=0):
        """
        Set the parameters for vascular construction
        and simulation for SimVascular python interface.
        """
        self.set_geometry_options()
        self.set_loft_options()
        self.set_solid_options(num_caps)
        self.set_mesh_options()
        self.set_pipeline_steps(gui=gui)
        self.set_directories()
        if len(time) > 2:
            self.simulation_parameters(fourier_mode=10,period=time[-1],bct_point_number=2*len(time),distal_resistance=distal_resistance)
        else:
            self.simulation_parameters(distal_resistance=distal_resistance)
        self.flow_parameters(time=time,flow=flow)
        self.rom_parameters()

    def set_geometry_options(self,number_samples=50,use_distance_alignment_method=True):
        geometry_options = {'number_samples':                number_samples,
                            'use_distance_alignment_method': use_distance_alignment_method}
        self.geometry_options = geometry_options
        return
    def set_loft_options(self,u_knot_span_type=None,v_parametric_span_type=None,
                         u_parametric_span_type=None,v_knot_span_type=None,
                         v_degree=None,u_degree=None,boundary_face_angle=45):
        loft_param = [u_knot_span_type,v_parametric_span_type,
                      u_parametric_span_type,v_knot_span_type,
                      v_degree,u_degree]
        loft_param = [param is None for param in loft_param]
        if all(loft_param):
            loft_options = {'default':             True,
                            'changes':               '',
                            'angle':boundary_face_angle}
        else:
            loft_options = {'default': False,
                            'changes':    '',
                            'angle':boundary_face_angle}
            # Check Knot Span Values
            if u_knot_span_type is not None:
                if u_knot_span_type not in ['derivative','equal','average']:
                    print('Invalid knot span type\n\nMust be one of:\n  "derivative"\n  "equal"\n  "average"\n')
                    print('Will default to "derivative" type.')
                else:
                    loft_options['changes'] += 'options.u_knot_span_type = {}\n'.format(u_knot_span_type)
            if v_knot_span_type is not None:
                if v_knot_span_type not in ['derivative','equal','average']:
                    print('Invalid knot span type\n\nMust be one of:\n  "derivative"\n  "equal"\n  "average"\n')
                    print('Will default to "average" type.')
                else:
                    loft_options['changes'] += 'options.v_knot_span_type = {}\n'.format(v_knot_span_type)
            # Check Parametric Span Values
            if v_parametric_span_type is not None:
                if v_parametric_span_type not in ['centripetal','equal','chord']:
                    print('Invalid knot span type\n\nMust be one of:\n  "centripetal"\n  "equal"\n  "chord"\n')
                    print('Will default to "chord" type.')
                else:
                    loft_options['changes'] += 'options.v_parametric_span_type = {}\n'.format(v_parametric_span_type)
            if u_parametric_span_type is not None:
                if u_parametric_span_type not in ['centripetal','equal','chord']:
                    print('Invalid knot span type\n\nMust be one of:\n  "centripetal"\n  "equal"\n  "chord"\n')
                    print('Will default to "centripetal" type.')
                else:
                    loft_options['changes'] += 'options.u_parametric_span_type = {}\n'.format(u_parametric_span_type)
            if v_degree is not None:
                if not isinstance(v_degree,int):
                    print('v_degree must be an integer')
                    print('Will default to v_degree = 2')
                else:
                    loft_options['changes'] += 'options.v_degree = {}\n'.format(v_degree)
            if u_degree is not None:
                if not isinstance(u_degree,int):
                    print('v_degree must be an integer')
                    print('Will default to u_degree = 2')
                else:
                    loft_options['changes'] += 'options.u_degree = {}\n'.format(u_degree)
        self.loft_options = loft_options
        return
    def set_solid_options(self,num_caps,minimum_face_cells=200,hmin=0.02,hmax=0.02,
                          face_edge_size=0.02,boundary_face_angle=45):
        #Check input types
        if not isinstance(minimum_face_cells,int):
            print('minimum_face_cells must be a positive int > 0')
            print('Using default value')
            minimum_face_cells=200
        if not isinstance(hmin,float):
            print('hmin must be a float < 1')
            print('Using default value')
            hmin=0.02
        if not isinstance(hmax,float):
            print('hmax must be a float hmin =< hmax < 1')
            print('Using default value')
            hmax=0.02
        if not isinstance(face_edge_size,float):
            print('face_edge_size must be a float < 1')
            print('Using default value')
            face_edge_size=0.02
        if boundary_face_angle > 90 or boundary_face_angle < 0:
            print('boundary_face_angle must be greater than 0 and less than 90')
            print('Using default value')
            boundary_face_angle = 45
        solid_options = {'minimum_face_cells':  minimum_face_cells,
                         'hmin':                             hmin,
                         'hmax':                             hmax,
                         'face_edge_size':          face_edge_size,
                         'angle':              boundary_face_angle,
                         'num_caps':                      num_caps}
        self.solid_options = solid_options
        return
    def set_mesh_options(self,global_edge_size=0.01,surface_mesh_flag=True,
                         volume_mesh_flag=True,no_merge=False,optimization_level=5,
                         minimum_dihedral_angle=18.0):
        if not isinstance(global_edge_size,float):
            print('global_edge_size must be a float less than 1')
            print('Using default: global_edge_size = 0.01')
            global_edge_size=0.01
        if not isinstance(surface_mesh_flag,bool):
            print('surface_mesh_flag must be boolean logic')
            print('Using default: surface_mesh_flag = True')
            surface_mesh_flag = True
        if not isinstance(volume_mesh_flag,bool):
            print('volume_mesh_flag must be boolean logic')
            print('Using default: volume_mesh_flag = True')
            volume_mesh_flag = True
        if not isinstance(no_merge,bool):
            print('no_merge must be boolean logic')
            print('Using default: no_merge = False')
            no_merge = False
        if not isinstance(optimization_level,int):
            print('optimization_level must be an integer between 0 and 7, inclusive')
            print('Using default: optimization_level = 5')
            optimization_level = 5
        elif optimization_level > 7 or optimization_level < 0:
            print('optimization_level must be an integer between 0 and 7, inclusive')
            print('Using default: optimization_level = 5')
            optimization_level = 5
        if not isinstance(minimum_dihedral_angle,float):
            print('minimum_dihedral_angle must be a float')
            print('Using default: minimum_dihedral_angle = 18')
            minimum_dihedral_angle = 18
        mesh_options = {'global_edge_size':            global_edge_size,
                        'surface_mesh_flag':          surface_mesh_flag,
                        'volume_mesh_flag':            volume_mesh_flag,
                        'no_merge':                            no_merge,
                        'optimization_level':        optimization_level,
                        'minimum_dihedral_angle':minimum_dihedral_angle}
        self.mesh_options = mesh_options
        return
    def set_pipeline_steps(self,files='all',gui=False):
        file_types = ['path','contour','solid',
                      'mesh','simulation','rom']
        # Check parameter values
        if not isinstance(files,str):
            print('{files} parameter must be a str')
            return
        if not isinstance(gui,bool):
            print('{gui} parameter must be True or False')
            return
        # Assign Model Constructor Values
        file_constructor = {'path':        True,
                            'contour':     True,
                            'solid':       True,
                            'mesh':        True,
                            'simulation':  True,
                            'rom':         True,
                            'gui':          gui}
        if files == 'all':
            pass
        else:
            if files in file_types:
               file_index = file_types.index(files)
               for i,f in enumerate(file_types):
                   if i <= file_index:
                       pass
                   else:
                       file_constructor[f] = False
        self.file_constructor = file_constructor
        return

    def set_directories(self,outdir=None):
        if isinstance(outdir,type(None)):
            outdir = os.getcwd()
        if isinstance(outdir,str):
            if not os.path.isdir(outdir):
                print('{} either does not exist or is not accessible'.format(outdir))
                return
            else:
                pass
        else:
            print('{outdir} should be a str argument if provided')
            return
        # Assigning Directory Constructor Values
        directory_constructor = {'main_folder':          None,
                                 'create_main_folder':  False,
                                 'data_folder':          None,
                                 'mesh_complete_folder': None,
                                 'mesh_surfaces_folder': None,
                                 'postprocessing_folder':None}
        if not os.path.isdir(outdir+'/CCO_DATA'):
            directory_constructor['create_main_folder'] = True
        else:
            directory_constructor['create_main_folder'] = False
        directory_constructor['main_folder'] = outdir+os.sep+'CCO_DATA'
        number = len(glob.glob(directory_constructor['main_folder']+os.sep+'CCO_*'))
        directory_constructor['data_folder'] = directory_constructor['main_folder']+os.sep+'simulation_data{}'.format(number)
        directory_constructor['mesh_complete_folder'] = directory_constructor['data_folder']+os.sep+'mesh-complete'
        directory_constructor['mesh_surfaces_folder'] = directory_constructor['mesh_complete_folder']+os.sep+'mesh-surfaces'
        directory_constructor['centerline_folder'] = directory_constructor['data_folder']+os.sep+'centerlines'
        directory_constructor['postprocessing_folder'] = directory_constructor['data_folder']+os.sep+'post'
        directory_constructor['rom_folder'] = directory_constructor['data_folder']+os.sep+'rom'
        directory_constructor['rom_0D_folder'] = directory_constructor['rom_folder']+os.sep+'zeroD'
        directory_constructor['rom_1D_folder'] = directory_constructor['rom_folder']+os.sep+'oneD'
        directory_constructor['main_folder']           = directory_constructor['main_folder'].replace('\\','\\\\')
        directory_constructor['data_folder']           = directory_constructor['data_folder'].replace('\\','\\\\')
        directory_constructor['mesh_complete_folder']  = directory_constructor['mesh_complete_folder'].replace('\\','\\\\')
        directory_constructor['mesh_surfaces_folder']  = directory_constructor['mesh_surfaces_folder'].replace('\\','\\\\')
        directory_constructor['centerline_folder']     = directory_constructor['centerline_folder'].replace('\\','\\\\')
        directory_constructor['postprocessing_folder'] = directory_constructor['postprocessing_folder'].replace('\\','\\\\')
        directory_constructor['rom_folder']            = directory_constructor['rom_folder'].replace('\\','\\\\')
        directory_constructor['rom_0D_folder']         = directory_constructor['rom_0D_folder'].replace('\\','\\\\')
        directory_constructor['rom_1D_folder']         = directory_constructor['rom_1D_folder'].replace('\\','\\\\')
        self.directory_constructor = directory_constructor
        self.number = number
        self.filename = directory_constructor['main_folder']+os.sep+'CCO_{}.py'.format(number)
        self.filename = self.filename.replace('\\','\\\\')
        self.name = 'CCO_{}'.format(number)
        return

    def simulation_parameters(self,svpre_name='cco',fluid_density=1.06,
                              fluid_viscosity=0.04, initial_pressure=0,
                              initial_velocity=[0.0001,0.0001,0.0001],
                              bct_analytical_shape='parabolic',
                              inflow_file='inflow',
                              period=1,bct_point_number=2,
                              fourier_mode=1,pressures=0,
                              svsolver='solver',number_timesteps=200,
                              timestep_size=0.02,number_restarts=50,
                              number_force_surfaces=1,
                              surface_id_force_calc=1,
                              force_calc_method='Velocity Based',
                              print_avg_solution=True,
                              print_error_indicators=False,
                              varying_time_from_file=True,
                              step_construction='0 1 0 1',
                              pressure_coupling='Implicit',
                              backflow_stabilization=0.2,
                              residual_control=True,
                              residual_criteria=0.01,
                              minimum_req_iter=2,
                              svLS_type='NS',num_krylov=100,
                              num_solves_per_left=1,
                              tolerance_momentum=0.05,
                              tolerance_continuity=0.4,
                              tolerance_svLS_NS=0.4,
                              max_iter_NS=2,max_iter_momentum=4,
                              max_iter_continuity=400,
                              time_integration_rule='Second Order',
                              time_integration_rho=0.5,
                              flow_advection_form='Convective',
                              quadrature_interior=2,
                              quadrature_boundary=3,procs=24,
                              svpost='cco',start=0,vtu=True,
                              vtp=False,vtkcombo=False,
                              all_arg=True,wss=False,
                              sim_units_mm=False,
                              sim_units_cm=True,
                              global_edge_size=0.01,
                              distal_resistance=0):
        # Assign PreSolver Constructor Values
        presolver_constructor = {'name':                           svpre_name,
                                 'fluid_density':               fluid_density,
                                 'fluid_viscosity':           fluid_viscosity,
                                 'initial_pressure':         initial_pressure,
                                 'initial_velocity':         initial_velocity,
                                 'bct_analytical_shape': bct_analytical_shape,
                                 'inflow_file':        inflow_file+'_3d.flow',
                                 'period':                             period,
                                 'bct_point_number':         bct_point_number,
                                 'fourier_mode':                 fourier_mode,
                                 'pressures':                       pressures}
        # Assign Solver Constructor Values
        solver_constructor = {'name':                                svsolver,
                              'fluid_density':                  fluid_density,
                              'fluid_viscosity':              fluid_viscosity,
                              'number_timesteps':            number_timesteps,
                              'timestep_size':                  timestep_size,
                              'number_restarts':              number_restarts,
                              'number_force_surfaces':  number_force_surfaces,
                              'surface_id_force_calc':  surface_id_force_calc,
                              'force_calc_method':          force_calc_method,
                              'print_avg_solution':        print_avg_solution,
                              'print_error_indicators':print_error_indicators,
                              'varying_time_from_file':varying_time_from_file,
                              'step_construction':          step_construction,
                              'pressure_coupling':          pressure_coupling,
                              'backflow_stabilization':backflow_stabilization,
                              'residual_control':            residual_control,
                              'residual_criteria':          residual_criteria,
                              'minimum_req_iter':            minimum_req_iter,
                              'svLS_type':                          svLS_type,
                              'num_krylov':                        num_krylov,
                              'num_solves_per_left':      num_solves_per_left,
                              'tolerance_momentum':        tolerance_momentum,
                              'tolerance_continuity':    tolerance_continuity,
                              'tolerance_svLS_NS':          tolerance_svLS_NS,
                              'max_iter_NS':                     max_iter_NS,
                              'max_iter_momentum':          max_iter_momentum,
                              'max_iter_continuity':      max_iter_continuity,
                              'time_integration_rule':  time_integration_rule,
                              'time_integration_rho':    time_integration_rho,
                              'flow_advection_form':      flow_advection_form,
                              'quadrature_interior':      quadrature_interior,
                              'quadrature_boundary':      quadrature_boundary,
                              'global_edge_size':            global_edge_size,
                              'distal_resistance':          distal_resistance}
        indir = self.directory_constructor['data_folder']+'/{}-procs_case'.format(procs)
        outdir = self.directory_constructor['postprocessing_folder']
        if sim_units_mm and sim_units_cm:
            print('Must pick one unit scale. Not both.')
            print('Exiting without assigning postprocessing parameters.')
            return
        postsolver_constructor = {'name':                              svpost,
                                  'start':                              start,
                                  'stop':                    number_timesteps,
                                  'incr':                     number_restarts,
                                  'vtu':                                  vtu,
                                  'vtp':                                  vtp,
                                  'vtkcombo':                        vtkcombo,
                                  'all':                              all_arg,
                                  'wss':                                  wss,
                                  'indir':                              indir,
                                  'sim_units_mm':                sim_units_mm,
                                  'sim_units_cm':                sim_units_cm,
                                  'outdir':                            outdir}
        self.presolver_constructor   = presolver_constructor
        self.solver_constructor      = solver_constructor
        self.postsolver_constructor  = postsolver_constructor
        return
    def flow_parameters(self,inflow_file='inflow',time=[0,1],flow=[1,1]):
        flow_constructor = {'inflow_file':                        inflow_file,
                            'time':                                      time,
                            'flow':                                      flow}
        self.flow_constructor = flow_constructor

    def rom_parameters(self,inflow='inflow_rom.flow',zeroD=True,oneD=True,time_step=0.001,
                       number_time_steps=1000):
        rom0D_constructor = {'create': zeroD,
                             'model_order': 0,
                             'time_step': time_step,
                             'number_time_steps': number_time_steps,
                             'inflow_file':  inflow}

        rom1D_constructor = {'create': oneD,
                             'model_order': 1,
                             'time_step': time_step,
                             'number_time_steps': number_time_steps,
                             'inflow_file':   inflow}

        self.rom_constructor = {'0D': rom0D_constructor,
                                '1D': rom1D_constructor}
        return



import svcco
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import pyvista as pv
from tqdm import tqdm
import tetgen
from scipy import stats
import pymeshfix
import pickle

q = 4
resolution = 120


cu = pv.Cube(x_length=3.72,y_length=3.72,z_length=3.72).triangulate().subdivide(5)
cube = svcco.surface()
cube.set_data(cu.points,cu.point_normals)
cube.solve()
cube.build(q=q,resolution=resolution)
print('cube constructed')

heart = svcco.surface()
heart_points = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_points_unique.csv',delimiter=',')
heart_normals = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_normals_unique.csv',delimiter=',')
heart.set_data(heart_points,heart_normals)
heart.solve()
heart.build(q=4,resolution=120,k=2,buffer=5)
print('heart constructed')


disk = pv.Disc(inner=2.8,outer=4,r_res=20,c_res=100)
cyl  = disk.extrude([0,0,2],capping=True).triangulate()
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl)
cyl  = cyl.subdivide(2)
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl,hausd=0.005)
cyl  = cyl.compute_normals(auto_orient_normals=True,feature_angle=90)
cylinder = svcco.surface()
cylinder.set_data(cyl.points,cyl.point_normals)
cylinder.solve()
cylinder.build(q=q,resolution=resolution)
print('cylinder constructed')

left_gyrus   = "D:\\Tree\\Tree_8-0\\brain_testing\\FJ3801_BP58201_FMA72658_Left inferior frontal gyrus.obj"
gyrus_no_scale = pv.read(left_gyrus)
sf = (heart.volume/gyrus_no_scale.volume)**(1/3)
gyrus_scaled = gyrus_no_scale.scale([sf,sf,sf])
left_gyrus_scaled = "left_gyrus_scaled.vtp"
gyrus_scaled.save(left_gyrus_scaled)
gyrus = svcco.surface()
gyrus.load(left_gyrus_scaled)
gyrus.solve()
gyrus.build(q=q,resolution=resolution,buffer=5)
print('gyrus constructed')

###########################################
# Test function
###########################################

def perfusion_volumes(tree,subdivisions=0,tet=None):
    terminals = tree.data[tree.data[:,15]<0,:]
    terminals = terminals[terminals[:,16]<0,:]
    territory_id = []
    territory_volumes = np.zeros(len(terminals))
    if tet is None:
        if subdivisions > 0:
            surf = tree.boundary.tet.grid.extract_surface()
            surf = surf.triangulate()
            surf = surf.subdivide(subdivisions)
            surf = svcco.utils.remeshing.remesh.remesh_surface(surf)
            surf = svcco.utils.remeshing.remesh.remesh_surface(surf)
            meshfix = pymeshfix.MeshFix(surf)
            meshfix.repair(verbose=True)
            meshfix.repair(verbose=True)
            surf = meshfix.mesh
            #surf = surf.subdivide(subdivisions)
            #surf = svcco.utils.remeshing.remesh.remesh_surface(surf)
            #surf = svcco.utils.remeshing.remesh.remesh_surface(surf)
            #meshfix = pymeshfix.MeshFix(surf)
            #meshfix.repair(verbose=True)
            #meshfix.repair(verbose=True)
            #surf = meshfix.mesh
            tet  = tetgen.TetGen(surf)
            tet.tetrahedralize()
        else:
            tet = tree.boundary.tet
    vol = tet.grid.compute_cell_sizes().cell_data['Volume']
    for idx, cell_center in tqdm(enumerate(tet.grid.cell_centers().points),desc='Calculating Perfusion Territories'):
        #for idx, cell_center in enumerate(tree.boundary.tet.grid.cell_centers().points):
        closest_terminal = np.argmin(np.linalg.norm(cell_center.reshape(1,-1)-terminals[:,3:6],axis=1))
        territory_id.append(closest_terminal)
        territory_volumes[closest_terminal] += vol[idx]
    territory_id = np.array(territory_id)
    tet.grid['perfusion_territory_id'] = territory_id
    return territory_id,territory_volumes/np.sum(territory_volumes),tet

def test(surf_object,size=1000,restarts=50):
    t = svcco.tree()
    t.set_boundary(surf_object)
    delaunay = surf_object.pv_polydata.delaunay_3d()
    convexity = surf_object.volume/delaunay.volume
    if convexity > 0.95:
        t.convex = True
    t.set_root()
    add_amount = size//restarts
    VOLS  = []
    COUNT = []
    _,vols,_ = perfusion_volumes(t)
    for i in range(restarts):
        t.n_add(add_amount)
        if i == 0:
            _,vols,tet = perfusion_volumes(t)
        else:
            _,vols,tet = perfusion_volumes(t,tet=tet)
        VOLS.append(vols)
        if i == 0:
            COUNT.append(add_amount)
        else:
            COUNT.append(COUNT[-1]+add_amount)
    return VOLS,COUNT

def test_heart(size=1000,restarts=50):
    t = svcco.tree()
    t.set_boundary(heart)
    t.set_root()
    add_amount = size//restarts
    VOLS  = []
    COUNT = []
    _,vols,_ = perfusion_volumes(t)
    #VOLS.append(vols)
    #COUNT.append(1)
    for i in range(restarts):
        t.n_add(add_amount)
        if i == 0:
            _,vols,tet = perfusion_volumes(t)
        else:
            _,vols,tet = perfusion_volumes(t,tet=tet)
        VOLS.append(vols)
        if i == 0:
            COUNT.append(add_amount)
        else:
            COUNT.append(COUNT[-1]+add_amount)
    return VOLS,COUNT

def test_cylinder(size=1000,restarts=50):
    t = svcco.tree()
    t.set_boundary(c)
    t.convex = True
    t.set_root()
    add_amount = size//restarts
    VOLS  = []
    COUNT = []
    _,vols,_ = perfusion_volumes(t)
    #VOLS.append(vols)
    #COUNT.append(1)
    for i in range(restarts):
        t.n_add(add_amount)
        if i == 0:
            _,vols,tet = perfusion_volumes(t)
        else:
            _,vols,tet = perfusion_volumes(t,tet=tet)
        VOLS.append(vols)
        if i == 0:
            COUNT.append(add_amount)
        else:
            COUNT.append(COUNT[-1]+add_amount)
    return VOLS,COUNT

def test_gyrus(size=1000,restarts=50):
    t = svcco.tree()
    t.set_boundary(gyrus)
    t.set_root()
    add_amount = size//restarts
    VOLS  = []
    COUNT = []
    _,vols,_ = perfusion_volumes(t)
    #VOLS.append(vols)
    #COUNT.append(1)
    for i in range(restarts):
        t.n_add(add_amount)
        if i == 0:
            _,vols,tet = perfusion_volumes(t)
        else:
            _,vols,tet = perfusion_volumes(t,tet=tet)
        VOLS.append(vols)
        if i == 0:
            COUNT.append(add_amount)
        else:
            COUNT.append(COUNT[-1]+add_amount)
    return VOLS,COUNT

def results(test_list,size=100,restarts=100,bins=50,repeat=1):
    fig,ax = plt.subplots(nrows=2,ncols=2)
    #ax = ax.flatten()
    #cw = plt.get_cmap('coolwarm')
    norm = mpl.colors.Normalize(vmin=0, vmax=size+1)
    cmap = mpl.cm.ScalarMappable(norm=norm, cmap=mpl.cm.coolwarm)
    cmap.set_array([])
    DATA = {'volumes':[],'tree_size':[]}
    for i in range(len(test_list)):
        vols,sizes = test(test_list[i],size=size,restarts=restarts)
        DATA['volumes'].append(vols)
        DATA['tree_size'].append(sizes)
        idx = i//2
        jdx = i%2
        for j in range(len(vols)):
            #color_scale = (sizes[j]/size)*255
            x_min = min(vols[-1])
            x_max = max(vols[-1])
            x_range = np.linspace(x_min,x_max)
            clean_vols = vols[j][vols[j]>0]
            freq,edges = np.histogram(clean_vols,range=(0,1),bins=bins)
            centers = 0.5*(edges[1:]+edges[:-1])
            width   = edges[1]-edges[0]
            #if j == 0:
            #    ax[idx][jdx].bar(centers,freq/len(vols[j]),width=width,color='blue')
            #else:
            ax[idx][jdx].bar(centers,freq/len(clean_vols),width=width,alpha=0.4,color=cmap.to_rgba(len(vols[j])))
            #ax[idx][jdx].hist(vols[j],bins=50,density=True,alpha=0.25,color=cw(color_scale),range = (x_min,x_max))
            #kde = stats.gaussian_kde(vols[j])
            #ax[idx][jdx].plot(x_range,kde(x_range)/10,color=cw(color_scale),alpha=0.75)
    #sm = plt.cm.ScalarMappable(cmap='coolwarm')
    #sm.set_array(range(0,size))
    if repeat > 1:
        for j in range(repeat-1):
            for i in range(len(test_list)):
                vols,sizes = test(test_list[i],size=size,restarts=restarts)
                for k in range(len(vols)):
                    DATA['volumes'][i][k].extend(vols[k])


    fig.colorbar(cmap,ax=ax.ravel().tolist(),label="Number of Terminals")
    fig.savefig('perfusion-{}_num_vessels-{}_restarts-{}_num_bins-{}.svg'.format(len(test_list),size,restarts,bins),format='svg')
    return fig,ax,DATA



def get_hist_data(data,hist_range=(0,1),bins=50):
    data['frequency'] = []
    data['centers']   = []
    data['widths']    = []
    for idx in range(len(data['volumes'])):
        temp_freq  = []
        temp_cent  = []
        temp_width = []
        for jdx in range(len(data['volumes'][idx])):
            vols = data['volumes'][idx][jdx]
            clean_vols = vols[vols>0]
            freq,edges = np.histogram(clean_vols,range=hist_range,bins=bins)
            centers = 0.5*(edges[1:]+edges[:-1])
            temp_freq.append(freq/len(clean_vols))
            temp_cent.append(centers)
            temp_width.append(edges[1]-edges[0])
        data['frequency'].append(temp_freq)
        data['centers'].append(temp_cent)
        data['widths'].append(temp_width)
    return data

f,a,data = results([cube,cylinder,heart,gyrus])
file = open('perfusion_data.pkl','wb+')
pickle.dump(data,file)
file.close()



import matplotlib.pyplot as plt
import numpy as np

def reynolds(*args,**kwargs):
    object = args[0]
    mean_velocity = object.data[:,22]/(object.data[:,21]**2*np.pi)
    length = object.data[:,21]*2
    dynamic_viscosity = object.parameters['nu']
    density = object.parameters['rho']
    re = (density*mean_velocity*length)/dynamic_viscosity
    return re



from tqdm import tqdm
from .weingarten_map import *
import numpy as np

def sampling(points,angle=0,max_local_size=20,min_local_size=10,normals=None,l=0.5,quiet=True):
    """
    """
    # want to sample points that have high absolute
    # gaussian curvature because the local behavior
    # deviates from a parabolic shape thus more
    # patches will be needed to better approximate.
    curvature,idx,KDT = estimate_gaussian_curvature(points,quiet=quiet)
    abs_curvature = abs(curvature)
    sorted_abs_curvature_idx = abs_curvature.flatten().argsort().tolist()
    patches = []
    patch_idx = []
    centers = []
    total_length = len(sorted_abs_curvature_idx)
    dist, idx = KDT.query(points,k=max_local_size+1)
    if quiet:
        while len(sorted_abs_curvature_idx) > 0:
            center = sorted_abs_curvature_idx[0]
            count = 0
            patch_idxs = []
            for k in idx[center,:]:
                #if abs(abs_curvature[center] - abs_curvature[k]) > 0.1:
                #    pass
                if np.isclose(np.dot(normals[center],normals[k]),1) and np.all(points[center]==points[k]):
                    patch_idxs.append(count)
                elif np.dot(normals[center],normals[k]) <= angle or abs(abs_curvature[center] - abs_curvature[k]) > 0.2:
                    if len(patch_idxs) > min_local_size:
                        break
                else:
                    patch_idxs.append(count)
                #if np.dot(normals[center],normals[k]) <= angle or abs(abs_curvature[center] - abs_curvature[k]) > 0.2:#np.all(points[center]==points[k]):
                #    if len(patch_idxs) > min_local_size:
                #        break
                #else:
                #    patch_idxs.append(count)
                #    #pass
                count += 1
            #limit = count//3
            if len(patch_idxs) < round(min_local_size*0.5):
                limit = len(patch_idxs)
            else:
                limit = round(len(patch_idxs)*l)
            #if len(patch_idxs) >= min_local_size:
            #    #if limit == 0:
            #    #       limit += 1
            #    #if limit >= 1:
            #    patches.append(points[idx[center,patch_idxs],:])
            #    patch_idx.append(idx[center,patch_idxs])
            #    centers.append(center)
            #    for i in idx[center,patch_idxs]:
            #        if i in sorted_abs_curvature_idx:
            #            sorted_abs_curvature_idx.remove(i)
            #else:
            #    #if idx[center,patch_idxs[0]] in sorted_abs_curvature_idx:
            #    #    sorted_abs_curvature_idx.remove(idx[center,patch_idxs[0]])
            #    sorted_abs_curvature_idx.remove(center)
            patches.append(points[idx[center,patch_idxs],:])
            patch_idx.append(idx[center,patch_idxs])
            centers.append(center)
            #print(points[idx[center,patch_idxs],:])
            for i in idx[center,patch_idxs[:limit]]:

                if i in sorted_abs_curvature_idx:
                    sorted_abs_curvature_idx.remove(i)
            #if limit-1 > 0:
            #    pbar.update(limit-1)
        return patches,patch_idx,KDT,centers
    with tqdm(total=total_length,desc='Determining Curvature  ') as pbar:
        while len(sorted_abs_curvature_idx) > 0:
            center = sorted_abs_curvature_idx[0]
            count = 0
            patch_idxs = []
            for k in idx[center,:]:
                #if abs(abs_curvature[center] - abs_curvature[k]) > 0.1:
                if np.dot(normals[center],normals[k]) <= angle:
                    if count > min_local_size:
                        break
                else:
                    patch_idxs.append(count)
                    #pass
                count += 1
            #limit = count//3
            limit = round(len(patch_idxs)*l)
            if len(patch_idxs) >= min_local_size:
                #if limit == 0:
                #	limit += 1
                #if limit >= 1:
                patches.append(points[idx[center,patch_idxs],:])
                patch_idx.append(idx[center,patch_idxs])
                centers.append(center)
                for i in idx[center,patch_idxs]:
                    if i in sorted_abs_curvature_idx:
                        sorted_abs_curvature_idx.remove(i)
                        pbar.update(1)
            else:
                #if idx[center,patch_idxs[0]] in sorted_abs_curvature_idx:
                #    sorted_abs_curvature_idx.remove(idx[center,patch_idxs[0]])
                sorted_abs_curvature_idx.remove(center)
                pbar.update(1)
            #if limit-1 > 0:
            #    pbar.update(limit-1)
    return patches,patch_idx,KDT,centers



r'''
# Automated Vascularization & Fluid Simulation for Engineered Tissues

`svcco` falls under the open-source SimVascular project which seeks to
understand and quantify hemodynamics in health and disease.
This python module creates synthetic microvascular networks for given
perfusion territories, a common problem when creating engineered tissues.

## Installing SVCCO

Conda install
```bash
# Best practice to use an environment rather than installing at the base
# By default we will specify the python environment to 3.7
# however 3.8 and 3.9 are valid
conda create -n svcco-env python=3.7
conda activate svcco-env
# becuase conda allows pip installing
pip install svcco
```

pip install

```bash
pip install svcco
```
'''
#from __future__ import annotations

__version__ = "0.5.30"

import traceback
import warnings
from pathlib import Path
from typing import overload

#from svcco import implicit, collision, branch_addition, sv_interface, forest_utils, utils

#@overload
#def svcco(*modules: Path | str, output_directory: None = None,
#          format):
#    pass
from . import implicit
from . import collision
from . import branch_addition
from . import sv_interface
from . import forest_utils
from . import utils

from .utils.remeshing import remesh
from .utils.gcode import gcode
from .utils.fluid_analysis import reynolds, wss
from .utils.gui import gui_helper

from .tree import tree, forest, perfusion_territory
from .implicit.implicit import surface
from .implicit.tests.bumpy_sphere import bumpy_sphere
from .implicit.tests.heart import *
from .implicit.visualize.visualize import plot_volume
#else:
#    implicit = _ModuleProxy('implicit')
#    collision = _ModuleProxy('collision')
#    branch_addition = _ModuleProxy('branch_addition')
#    sv_interface = _ModuleProxy('sv_interface')



##########
#UNIT TEST - COLLIDING CYLINDERS (SPHERE PROXIMITY)
##########

import numpy as np
from sphere_proximity import sphere_proximity

point_0 = np.array([0,0,-1])
point_1 = np.array([0,0,1])
point_2 = np.array([-1,0,0])
point_3 = np.array([1,0,0])

r0 = 0.1
r1 = 0.1

DATA = np.zeros((1,31))
EDGE = np.zeros(31)

l0 = np.linalg.norm(point_1-point_0)
l1 = np.linalg.norm(point_3-point_2)

DATA[0,0:3] = point_0
DATA[0,3:6] = point_1
EDGE[0:3] = point_2
EDGE[3:6] = point_3

DATA[0,20]  = l0
EDGE[20]    = l1

result = sphere_proximity(DATA,EDGE)

assert len(result) == 1, "Sphere proximity did not catch collision"



meta_info = """
// summary: G-code for Synthetic Vascular Printing
// authors : Zachary Sexton & Jessica Herrmann
// email  : zsexton@stanford.edu
"""

licence = """
// Copyright (c) Stanford University, The Regents of the University of
//               California, and others.
//
// All Rights Reserved.
//
// See Copyright-SimVascular.txt for additional details.
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject
// to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""

variables = """
// --------- Variable Declarations ----------- //
//    Variables                   Units            Description
DVAR $syringe_diameter            mm               diameter of syringe body
DVAR $nozzle_diameter             mm               diameter of nozzle outlet
DVAR $jog_speed                   mm/sec           movement speed while not printing
DVAR $overall_length              mm
DVAR $print_speed                 mm/sec           movement speed while printing
DVAR $extrusion_coeff             None             adjustment ratio between print speed and nozzle size
DVAR $plunger_rate                mm/sec           rate of syringe plunger depression
DVAR $x_start                                      starting x value
DVAR $y_start                                      starting y value
DVAR $z_start                                      starting z value
DVAR $x_offset                    mm               reference frame difference x direction
DVAR $y_offset                    mm               reference frame difference y direction
DVAR $z_offset                    mm               reference frame difference z direction
DVAR $initial_pump_distance       mm               starting distance for pump plunger
DVAR $initial_pump_rate           mm/sec           pump plunger rate

"""

initialize_variables = """
// --------- General Parameters ----------- //
$syringe_diameter      = 9.6        // 3BB Syringe = 9.55 mm, 10mL glass = 10.3mm
$jog_speed             = 10
$initial_pump_distance = 0.05
$initial_pump_rate     = 5

// --------- Tube Parameters ----------- //
$nozzle_diameter = 0.84
$print_speed     = 6
$extrusion_coeff = 1

// --------- Calculated Parameters ----------- //
$plunger_rate = $extrusion_coeff * $print_speed * $nozzle_diameter**2/$syringe_diameter**2

"""

stop_pump = """
Call StopExtrude\n
"""

start_pump = """
Call StartExtrude P$initial_pump_distance Q$initial_pump_rate R$plunger_rate
G90 F3
G108

"""

recalculate_plunger_rate = """
$extrusion_Boeff = {}
$plunger_rate = $extrusion_coeff * $print_speed * $nozzle_diameter**2 / $syringe_diameter**2

"""

branch_name = """
// Branch {}

"""

end_file = """
G90
G1 B60

M2

"""

start_extrude = """
DFS StartExtrude

	Enable Aa
	G91
	G1 Aa-$initial_pump_dist E$initial_pump_rate
	M104
	S$plunger_rate

EndDFS

"""

stop_extrude = """
DFS StopExtrude

	M5
	Enable Aa
	G91
	G1 Aa$initial_pump_dist E$initial_pump_rate

EndDFS
"""

def g_configure(x,y,z,precision=3):
    return "G90\nG92 X{} Y{} B{}\n\n".format(round(x,precision),round(y,precision),round(z,precision))

def g_coordinate(x,y,z,precision=3):
    return "G1 X{} Y{} B{}\n".format(round(x,precision),round(y,precision),round(z,precision))

def generate_gcode(paths,offset=[0,0,0]):
    file = open("test.gcode","w+")
    filetext = licence + meta_info + variables + initialize_variables
    for idx,path in enumerate(paths):
        filetext += branch_name.format(idx)
        filetext += g_configure(path[0][0],path[0][1],path[0][2])
        filetext += stop_pump
        filetext += start_pump
        for jdx,point in enumerate(path):
            filetext += g_coordinate(point[0],point[1],point[2])
        filetext += stop_pump
    filetext += end_file
    filetext += start_extrude
    filetext += stop_extrude
    file.write(filetext)



import numpy as np
import vtk
import os
from vtk.util import numpy_support
from ..utils.remeshing.remesh import remesh_surface
import pyvista as pv

def load3d(filename,subdivisions=0):
    vtk_reader = {'stl':vtk.vtkSTLReader,
                  'obj':vtk.vtkOBJReader,
                  'ply':vtk.vtkPLYReader,
                  'vtu':vtk.vtkXMLUnstructuredGridReader,
                  'vtp':vtk.vtkXMLPolyDataReader,
                  '3ds':vtk.vtk3DSImporter}

    ext = filename.rsplit('.')[1].lower()
    if ext not in vtk_reader.keys():
        print('Not a supported 3D file format')
        print('Supported Formats:\n{}'.format([supported_ext + '\n' for supported_ext in vtk_reader.keys() ]))
        return
    reader = vtk_reader[ext]()
    reader.SetFileName(filename)
    if ext == '3ds':
        reader.ComputeNormalsOn()
        reader.Update()
    elif ext == 'vtu':
        pass
    else:
        reader.Update()
        data = reader.GetOutput()
        points = data.GetPoints()
        points = points.GetData()
        points = numpy_support.vtk_to_numpy(points)
        data_normals = data.GetPointData()
        normals = data_normals.GetNormals()
        if subdivisions > 0:
            linearSubdivision = vtk.vtkLinearSubdivisionFilter()
            linearSubdivision.SetNumberOfSubdivisions(subdivisions)
            linearSubdivision.SetInputData(data)
            linearSubdivision.Update()
            data = linearSubdivision.GetOutput()
            points = data.GetPoints()
            points = points.GetData()
            points = numpy_support.vtk_to_numpy(points)
            normals_gen = vtk.vtkPolyDataNormals()
            normals_gen.SplittingOn()
            normals_gen.ComputeCellNormalsOff()
            normals_gen.ComputePointNormalsOn()
            normals_gen.SetInputData(data)
            normals_gen.Update()
            normal_polydata = normals_gen.GetOutput()
            normal_point_data = normal_polydata.GetPointData()
            normals = normal_point_data.GetNormals()
            normals = numpy_support.vtk_to_numpy(normals)
        else:
            #normals = numpy_support.vtk_to_numpy(normals)
            if normals is None:
                """
                normals_gen = vtk.vtkPolyDataNormals()
                normals_gen.SplittingOn()
                normals_gen.ComputeCellNormalsOff()
                normals_gen.ComputePointNormalsOn()
                normals_gen.SetInputData(data)
                normals_gen.Update()
                normal_polydata = normals_gen.GetOutput()
                normal_point_data = normal_polydata.GetPointData()
                normals = normal_point_data.GetNormals()
                normals = numpy_support.vtk_to_numpy(normals).tolist()
                points = []
                norms = []
                for idx in range(normal_polydata.GetNumberOfCells()):
                    tri = normal_polydata.GetCell(idx)
                    tri_points = numpy_support.vtk_to_numpy(tri.GetPoints().GetData()).tolist()
                    points.extend(tri_points)
                    for jdx in range(len(tri_points)):
                        norms.append(normals[idx])
                """
                obj = pv.PolyData(var_inp=data)
                obj = obj.compute_normals()
                #points = np.array(points)
                #normals = np.array(norms)
                points = obj.points
                normals = obj['Normals']
            else:
                normals = numpy_support.vtk_to_numpy(normals)
        upt,uid = np.unique(points,axis=0,return_index=True)
        points = points[uid]
        normals = normals[uid]
        #normals = numpy_support.vtk_to_numpy(normals)
        #Check and clean duplicate points
        #points,idx = np.unique(points,axis=0,return_index=True)
        #normals    = normals[idx]
        # later duplicate points will be allowed to accomodate C1
        # surfaces which will require splitting during VTK NORMAL
        # calculation. This will also have to make the splitting
        # and PU angle thresholds the same to allow for non-singluar
        # matricies.
        return points,normals,data

def load3d_pv(filename,subdivisions=0,remesh=True,max_points=10000,verbosity=0):
    mesh = pv.read(filename)
    if remesh:
        hausd = 0.01
        mesh = remesh_surface(mesh,hausd=hausd,verbosity=verbosity)
        previous = mesh.points.shape[0]
        while mesh.points.shape[0] > max_points and mesh.points.shape[0] <= previous:
            hausd += 0.01
            print('Target: {} | Current: {}'.format(max_points,mesh.points.shape[0]))
            previous = mesh.points.shape[0]
            mesh = remesh_surface(mesh,hausd=hausd,verbosity=verbosity)
        print('End Point Number: {}'.format(mesh.points.shape[0]))
    points  = mesh.points
    normals = mesh.point_normals
    upt,uid = np.unique(points,axis=0,return_index=True)
    points = points[uid]
    normals = normals[uid]
    mesh.save(os.getcwd()+os.sep+'temp.vtp')
    reader = vtk.vtkXMLPolyDataReader()
    reader.SetFileName(os.getcwd()+os.sep+'temp.vtp')
    reader.Update()
    mesh = reader.GetOutput()
    os.remove(os.getcwd()+os.sep+'temp.vtp')
    return points,normals,mesh



import numba as nb

@nb.jit(nopython=True,cache=True,nogil=True)
def add_flow(data,parent,flow):
    updated_flows = []
    while parent >= 0:
        data[parent,22] = flow + data[parent,22]
        updated_flows.append(parent)
        parent = int(data[parent,17].item())
    return updated_flows



import numpy as np
import os #remove before final
import time #not neccessary?
from .check import *
from .close import *
from .local_func_v6 import *
from ..collision.sphere_proximity import *
from ..collision.collision import *
from .add_bifurcation import *
from .sample_triad import *
from .triangle import * #might not need
from .basis import *
from scipy import interpolate
from scipy.spatial import KDTree
import matplotlib.pyplot as plt #remove before final
from .get_point import *
from mpl_toolkits.mplot3d import proj3d #remove before final
from .geodesic import extract_surface,geodesic
from ..implicit.visualize.visualize import show_mesh
from .finite_difference import finite_difference


def add_branch(data,boundary,edge_number=None,gamma=None,
               nu=None,Qterm=None,Pperm=None,Pterm=None,lam=None,k=None,
               mu=None,threshold_adjuster=0.9,
               threshold_exponent=0.5,max_skip=None,target_low=None,
               target_high=None,sub_division_map=None,sub_division_index=None,
               fraction=None,isforest=False,ishomogeneous=None,isconvex=None,
               isdirected=None,isclamped=None,radius_buffer=None):
    threshold_distance = (boundary.volume)**(1/3)/(k**threshold_exponent)
    threshold_distance += threshold_distance*0.5*np.random.random(1)
    success      = False
    if ishomogeneous:
        centers = (data[:,0:3] + data[:,3:6])/2
        while not success:
            total_attempts = 0
            attempt      = 0
            while total_attempts < 40:
                point,_ = boundary.pick(homogeneous=True)
                vessel, line_distances = close_exact(data,point)
                if sum(line_distances < threshold_distance) == 0 and min(line_distances) > 4*data[vessel[0],21]:
                    escape = False
                    for i in range(max_skip):
                        if data[vessel[i],22] <= 8*Qterm:
                            vessel = vessel[i]
                            escape = True
                            break
                    if escape:
                        break
                else:
                    if attempt < 10:
                        attempt += 1
                        total_attempts += attempt
                    else:
                        print('adjusting threshold')
                        threshold_distance *= threshold_adjuster
                        attempt = 0
            proximal = data[vessel,0:3]
            distal   = data[vessel,3:6]
            terminal = point
            num = 20
            triangle = get_local_points(data,vessel,terminal,num,isclamped)
            triangle = relative_length_constraint(triangle,proximal,distal,terminal,0.25)
            if not isconvex:
                triangle = boundary_constraint(triangle,boundary,2)
            triangle = angle_constraint(triangle,terminal,distal,-0.4,True)
            if len(triangle) == 0:
                continue
            triangle = angle_constraint(triangle,terminal,distal,0.75,False)
            if len(triangle) == 0:
                continue
            triangle = angle_constraint(triangle,terminal,proximal,0,False)
            if len(triangle) == 0:
                continue
            triangle = angle_constraint(triangle,distal,proximal,0,False)
            if len(triangle) == 0:
                continue
            if data[vessel,17] >= 0:
                p_vessel = int(data[vessel,17])
                daughter1_vector = -data[p_vessel,12:15]
                daughter2_vector = (triangle - proximal)/np.linalg.norm(triangle - proximal,axis=1).reshape(-1,1)
                angle = np.array([np.dot(daughter1_vector,daughter2_vector[i]) for i in range(len(daughter2_vector))])
                triangle = triangle[angle<0]
                if len(triangle) == 0:
                    continue
            #start = time.time()
            results  = fast_local_function(data,triangle,terminal,vessel,gamma,nu,Qterm,Pperm,Pterm)
            #b_time = time.time() - start
            volume       = np.pi*(results[0]**lam)*(results[1]**mu)
            idx          = np.argmin(volume)
            bif          = results[5][idx]
            #start = time.time()
            #fd_bif,fd_idx,fd_volume,fd_trial = finite_difference(data,triangle,terminal,vessel,gamma,nu,Qterm,Pperm,Pterm)
            #fd_time = time.time() - start
            no_collision = collision_free(data,results,idx,terminal,vessel,radius_buffer)
            #no_collision_fd = collision_free_fd(fd_trial,vessel)
            #print(no_collision)
            if no_collision:
                #print('LOCAL ACCEL: {}    FINITE DIFF: {}'.format(bif,fd_bif))
                #print(np.allclose(bif,fd_bif))
                #print('adding')
                data = np.vstack((data,np.zeros((2,data.shape[1]))))
                data,sub_division_index = add_bifurcation(data,None,bif,terminal,vessel,Qterm,
                                                          results[0],results[1],results[2],results[3],
                                                          np.array(results[6]),np.array(results[7]),
                                                          np.array(results[8]),np.array(results[9]),
                                                          np.array(results[10]),np.array(results[11]),
                                                          np.array(results[12]),np.array(results[13]),
                                                          results[14],results[15],idx,sub_division_map,
                                                          sub_division_index,True)
                #print('FINITE: {}    BINDING: {}'.format(fd_time,b_time))
                #data = fd_trial
                k += 2
                results = [data,k,sub_division_map,sub_division_index]
                if not isforest:
                     return data,k,sub_division_map,sub_division_index
                else:
                     new_terminal = data[-2,:]
                     new_sister   = data[-1,:]
                     new_parent   = data[vessel,:]
                     return new_terminal,new_sister,new_parent,results
            else:
                continue
                #return data,k,sub_division_map,sub_division_index
    pure_data    = data[data[:,-1]>-1]
    segment_data = data[data[:,-1]==-1]
    vessel       = np.random.choice(list(range(pure_data.shape[0])))
    vessel_path  = segment_data[segment_data[:,29].astype(int)==vessel]
    other_vessels= segment_data[segment_data[:,29].astype(int)!=vessel]
    if pure_data.shape[0] > 1:
        other_KDTree = KDTree((other_vessels[:,0:3]+other_vessels[:,3:6])/2)
    else:
        other_KDTree = None
    mesh,pa,cp,cd= boundary.mesh(vessel_path[1:,0:3],threshold_distance,threshold_distance//fraction,dive=0,others=other_KDTree)
    #show_mesh(mesh,np.array(pa),np.array(cp))
    mesh_poly    = extract_surface(mesh)
    proximal     = vessel_path[0,0:3]
    distal       = vessel_path[-1,3:6]
    terminal     = cp[-1]
    #results      = fast_local_function(pure_data,mesh,terminal,vessel,gamma,nu,Qterm,Pperm,Pterm)
    d_max        = max(data[vessel,20]*0.1,np.sum(cd)*0.1)
    mesh_truc    = mesh[np.linalg.norm(mesh-proximal,axis=1)>d_max]
    mesh_truc    = mesh_truc[np.linalg.norm(mesh_truc-distal,axis=1)>d_max]
    #show_mesh(mesh_truc,np.array(pa),np.array(cp))
    mesh_truc    = mesh_truc[np.linalg.norm(mesh_truc-terminal,axis=1)>d_max]
    #show_mesh(mesh_truc,np.array(pa),np.array(cp))
    results      = fast_local_function(pure_data,mesh_truc,terminal,vessel,gamma,nu,Qterm,Pperm,Pterm)
    volume       = np.pi*(results[0]**lam)*(results[1]**mu)
    idx          = np.argmin(volume)
    bif          = results[5][idx]
    proximal,distal,terminal = geodesic(mesh_poly,proximal,distal,terminal,bif)
    new_parent_size   = proximal['path'].shape[0]-1
    new_sister_size   = distal['path'].shape[0]-1
    new_terminal_size = terminal['path'].shape[0]-1
    ##########
    # Ensure continuity of vessels reapply start and end nodes
    ##########
    #terminal['path'][0,:] = cp[-1]
    #distal['path'][0,:]   = vessel_path[-1,3:6]
    proximal['path'][-1,:]  = vessel_path[0,0:3]
    new_parent        = np.zeros((new_parent_size,data.shape[1]))
    new_sister        = np.zeros((new_sister_size,data.shape[1]))
    new_terminal      = np.zeros((new_terminal_size,data.shape[1]))
    new_segment_data  = other_vessels
    new_terminal[:,0:3] = terminal['path'][:-1,:]
    new_terminal[:,3:6] = terminal['path'][1:,:]
    new_terminal[:,20]  = np.linalg.norm(new_terminal[:,0:3]-new_terminal[:,3:6])
    new_terminal[:,21]  = results[0][idx]*results[2][idx]
    new_terminal[:,29]  = np.float(pure_data.shape[0])
    new_terminal[:,-1]  = -1
    new_sister[:,0:3]   = distal['path'][:-1,:]
    new_sister[:,3:6]   = distal['path'][1:,:]
    new_sister[:,20]    = np.linalg.norm(new_sister[:,0:3] - new_sister[:,3:6])
    new_sister[:,21]    = results[0][idx]*results[3][idx]
    new_sister[:,29]    = np.float(pure_data.shape[0])+1
    new_sister[:,-1]    = -1
    new_parent[:,0:3]   = proximal['path'][:-1,:]
    new_parent[:,3:6]   = proximal['path'][1:,:]
    new_parent[:,20]    = np.linalg.norm(new_parent[:,0:3] - new_parent[:,3:6])
    new_parent[:,29]    = np.float(vessel)
    new_parent[:,-1]    = -1
    if vessel != 0:
        new_parent[:,21] = results[0][idx]*results[4][idx]
    else:
        new_parent[:,21] = results[0][idx]
    for i in range(max([new_terminal.shape[0],new_sister.shape[0],new_parent.shape[0]])):
        if i < new_terminal.shape[0]:
            basis(new_terminal,i)
        if i < new_sister.shape[0]:
            basis(new_sister,i)
        if i < new_parent.shape[0]:
            basis(new_parent,i)
    new_segments = np.vstack((new_segment_data,new_terminal,new_sister,new_parent))
    pure_data = np.vstack((pure_data,np.zeros((2,data.shape[1]))))
    data,sub_division_index = add_bifurcation(pure_data,new_segments,bif,cp[-1],vessel,Qterm,
                                              results[0],results[1],results[2],results[3],
                                              np.array(results[6]),np.array(results[7]),
                                              np.array(results[8]),np.array(results[9]),
                                              np.array(results[10]),np.array(results[11]),
                                              np.array(results[12]),np.array(results[13]),
                                              results[14],results[15],idx,sub_division_map,
                                              sub_division_index)
    k += 2
    if not isforest:
        return data,k,sub_division_map,sub_division_index
    else:
        results = [data,k,sub_division_map,sub_division_index]
        return new_terminal,new_sister,new_parent,results




make_results = """import pyvista as pv
import numpy as np
import os
from tqdm import tqdm

geom_data = np.genfromtxt("geom.csv",delimiter=",")
data = np.load("solver_0d_branch_results.npy",allow_pickle=True).item()

total = None
timepoints = []
min_pressure = np.inf
max_pressure = -np.inf
min_flow = np.inf
max_flow = -np.inf
min_wss = np.inf
max_wss = -np.inf
for idx in tqdm(range(len(data['time'])),desc="Building Timeseries ",position=0):
    #time_merge = None
    time = data['time'][idx]
    tmp_vessels = []
    for jdx in tqdm(range(len(data['flow'])),desc="Building Vessel Data",position=1,leave=False):
        vessel = list(data['flow'].keys())[jdx]
        start = geom_data[vessel,0:3]
        end   = geom_data[vessel,3:6]
        direction = (geom_data[vessel,3:6] - geom_data[vessel,0:3])/np.linalg.norm(geom_data[vessel,3:6] - geom_data[vessel,0:3])
        length    = geom_data[vessel,6]
        radius    = geom_data[vessel,7]
        number_segments = len(data['flow'][jdx]) - 1 #assume only 1 segment right now
        number_points   = len(data['flow'][jdx])
        for kdx in range(number_segments):
            center = (1/2)*direction*length + start
            vessel = pv.Cylinder(center=center,direction=direction,height=length,radius=radius)
            vessel = vessel.elevation(low_point=end,high_point=start,scalar_range=[data['pressure'][jdx][kdx+1][idx]/1333.33, data['pressure'][jdx][kdx][idx]/1333.33])
            if data['pressure'][jdx][kdx][idx]/1333.33 > max_pressure:
                max_pressure = data['pressure'][jdx][kdx][idx]/1333.33
            if data['pressure'][jdx][kdx+1][idx]/1333.33 < min_pressure:
                min_pressure = data['pressure'][jdx][kdx+1][idx]/1333.33
            vessel.rename_array('Elevation','Pressure [mmHg]',preference='point')
            vessel.cell_data['Flow [mL/s]'] = data['flow'][jdx][kdx][idx]
            re = (1.06*2*radius*((data['flow'][jdx][kdx][idx]/(np.pi*radius**2))/2))/0.04
            fd = 64/re
            wss = ((data['flow'][jdx][kdx][idx]/(np.pi*radius**2))/2)*fd*1.06
            vessel.cell_data['WSS [dyne/cm^2]']  = wss
            if max_flow < data['flow'][jdx][kdx][idx]:
                max_flow = data['flow'][jdx][kdx][idx]
            if min_flow > data['flow'][jdx][kdx][idx]:
                min_flow = data['flow'][jdx][kdx][idx]
            if max_wss < wss:
                max_wss = wss
            if min_wss > wss:
                min_wss = wss
            tmp_vessels.append(vessel)
        #if time_merge is None:
        #    time_merge = vessel
        #else:
        #    time_merge = time_merge.merge(vessel)
    time_merge = tmp_vessels[0].merge(tmp_vessels[1:])
    time_merge.field_data['time'] = time
    timepoints.append(time_merge)
    #if total is None:
    #    total = time_merge
    #else:
    #    total = total.merge(time_merge)
    if not os.path.isdir("timeseries"):
        os.mkdir("timeseries")

if not os.path.isdir("timeseries_for_pressure_gif"):
    os.mkdir("timeseries_for_pressure_gif")
if not os.path.isdir("timeseries_for_flow_gif"):
    os.mkdir("timeseries_for_flow_gif")
if not os.path.isdir("timeseries_for_wss_gif"):
    os.mkdir("timeseries_for_wss_gif")
total = timepoints[0].merge(timepoints[1:])
for i in tqdm(range(len(timepoints)),desc="Saving Timeseries",position=1):
    p = pv.Plotter(off_screen=True)
    p.add_mesh(timepoints[i],scalars='Pressure [mmHg]',clim=[round(min_pressure,4),round(max_pressure,4)],cmap="coolwarm")
    p.show(auto_close=True,screenshot=os.getcwd()+os.sep+"timeseries_for_pressure_gif"+os.sep+"time_point_{}.png".format(i))
    p = pv.Plotter(off_screen=True)
    p.add_mesh(timepoints[i],scalars='Flow [mL/s]',clim=[round(min_flow,4),round(max_flow,4)],cmap="GnBu")
    p.show(auto_close=True,screenshot=os.getcwd()+os.sep+"timeseries_for_flow_gif"+os.sep+"time_point_{}.png".format(i))
    p = pv.Plotter(off_screen=True)
    p.add_mesh(timepoints[i],scalars='WSS [dyne/cm^2]',clim=[round(min_wss,2),round(max_wss,2)],cmap="coolwarm")
    p.show(auto_close=True,screenshot=os.getcwd()+os.sep+"timeseries_for_wss_gif"+os.sep+"time_point_{}.png".format(i))
    timepoints[i].save(os.getcwd()+os.sep+"timeseries"+os.sep+"time_point_{}.vtp".format(i))

total.save(os.getcwd()+os.sep+"timeseries"+os.sep+"total.vtp")
"""



import numba as nb
import numpy as np

#@nb.jit(nopython=True,cache=True,nogil=True)
def update(data,gamma,nu):
    depth_max = np.amax(data[:,26])
    start = 0
    while int(depth_max) >= 0:
        edges = np.argwhere(data[:,26] == depth_max)
        depth_max = depth_max - 1
        for idx in edges:
            if data[idx,15] < 0 and data[idx,16] < 0:
                data[idx,25] = (8*nu/np.pi)*data[idx,20]
                data[idx,23] = 0.0
                data[idx,24] = 0.0
                data[idx,27] = 0.0
            else:
                left = int(data[idx, 15].item())
                right = int(data[idx, 16].item())
                if right == -1:
                    data[idx,25] = (8*nu/np.pi)*data[idx,20] + \
                                   data[left, 25]
                    data[idx,23] = 1
                    data[idx,24] = 0
                    data[idx,27] = (data[left,20]+data[left,27])
                else:
                    LR = ((data[left, 22]*
                           data[left, 25])/
                          (data[right, 22]*
                           data[right, 25])) ** (1/4)
                    lbif = (1+ LR ** (-gamma)) ** (-1/gamma)
                    rbif = (1+ LR ** (gamma)) ** (-1/gamma)
                    data[idx,25] = (8*nu/np.pi)*data[idx,20] + \
                                   ((lbif ** 4 / data[left, 25]) +
                                    (rbif ** 4 / data[right, 25])) ** -1
                    data[idx,23] = lbif
                    data[idx,24] = rbif
                    data[idx,27] = lbif**2 * (data[left,20]+data[left,27])+\
                                   rbif**2 * (data[right,20]+data[right,27])

nb.jit(nopython=True)
def naive(data,gamma,nu,start):
    if data[start,15] < 0 and data[start,16] < 0:
        data[start,25] = (8*nu/np.pi)*data[start,20]
        data[start,23] = 0.0
        data[start,24] = 0.0
        data[start,27] = 0.0
        return (8*nu/np.pi)*data[start,20]
    else:
        left = int(data[start,15].item())
        right = int(data[start,16].item())
        left_rr = naive(data,gamma,nu,left)
        right_rr = naive(data,gamma,nu,right)
        LR = ((data[left, 22]*
               left_rr)/
              (data[right, 22]*
               right_rr)) ** (1/4)
        lbif = (1+ LR ** (-gamma)) ** (-1/gamma)
        rbif = (1+ LR ** (gamma)) ** (-1/gamma)
        data[start,25] = (8*nu/np.pi)*data[start,20] + ((lbif ** 4 / left_rr) + (rbif ** 4 / right_rr)) ** -1
        data[start,23] = lbif
        data[start,24] = rbif
        data[start,27] = lbif**2 * (data[left,20]+data[left,27])+\
                         rbif**2 * (data[right,20]+data[right,27])
        return (8*nu/np.pi)*data[start,20] + ((lbif ** 4 / left_rr) + (rbif ** 4 / right_rr)) ** -1



# Code: Build Time Figure Between Convex and Nonconvex geometries

import numpy as np
from pickle import load
import matplotlib.pyplot as plt
from matplotlib import cm
from tqdm import tqdm
import os

###############################################
# Parameters to change for figure display
###############################################

tree_sizes = [100,1000,4000,7999]

#convex_directories    = [os.getcwd()+os.sep+'cube_times',os.getcwd()+os.sep+'sphere_times']
convex_directories = os.walk(os.getcwd()+os.sep+'convex_data')
convex_directories = [x[0] for x in convex_directories]
convex_directories = list(filter(lambda n: 'times' in n,convex_directories))

nonconvex_directories = os.walk(os.getcwd()+os.sep+'brain_data')
nonconvex_directories = [x[0] for x in nonconvex_directories]
nonconvex_directories = list(filter(lambda n: 'times' in n,nonconvex_directories))

linecolors = []
red = cm.Reds
for l in range(len(convex_directories)):
    linecolors.append(red(256-l))

blue = cm.Blues
for l in range(len(nonconvex_directories)):
    linecolors.append(blue(256-l))

convex_directories.extend(nonconvex_directories)
directories  = convex_directories


linestyles = [(0,())]*len(directories)

#for l in range(len(nonconvex_directories)):
#    if l == 0:
#        linestyles.append(tuple([0,tuple([])]))
#    else:
#        linestyles.append(tuple([0,tuple([3,l])]))

capsize = 5

###############################################
# Below is just code to unpack data
# there is no need to touch this.
###############################################

DATA = {}

for dirname in directories:
    files = os.listdir(dirname)
    DATA[dirname] = np.zeros((len(files),len(tree_sizes)))
    for i, fname in tqdm(enumerate(files),desc='loading {} directory'.format(dirname)):
        f = open(dirname+os.sep+fname,'rb')
        time_dat = load(f)
        for j, size in enumerate(tree_sizes):
            DATA[dirname][i,j] = sum(time_dat['total'][:size])
    DATA[dirname+'_mean'] = np.mean(DATA[dirname],axis=0)
    DATA[dirname+'_std']  = np.std(DATA[dirname],axis=0)

fig = plt.figure()
ax = fig.add_subplot(111,xscale='log',yscale='log')
ax.yaxis.set_ticks_position('both')
ax.xaxis.set_ticks_position('both')
ax.set_ylabel('Build Time (s)')
ax.set_xlabel('Tree Size (# Terminals)')
ax.grid(which='both')
for i, dirname in enumerate(directories):
    #ax.plot(tree_sizes,DATA[dirname+'_mean'],color=linecolors[i],linestyle=linestyles[i])
    ax.errorbar(tree_sizes,DATA[dirname+'_mean'],yerr=DATA[dirname+'_std'],
                capsize=capsize,ecolor=linecolors[i],linestyle=linestyles[i],
                color=linecolors[i])

plt.show()



import os
from copy import deepcopy
from .setup import setup as setup_file
from .paths import paths as path_file
from .contour import contour as contour_file
from .geometry import geometry as geometry_file
from .loft import loft as loft_file
from .solid import solid as solid_file
from .mesh import mesh as mesh_file
from .presolver import presolver as presolver_file
from .solver import solver as solver_file
from .centerline import centerline as centerline_file
from .rom import rom as rom_file
from .flow import flow as flow_file

def build(paths,radii,normals,options,setup_file=setup_file,
          path_file=path_file,contour_file=contour_file.split("\n"),
          geometry_file=geometry_file,loft_file=loft_file,
          solid_file=solid_file,mesh_file=mesh_file,
          presolver_file=presolver_file,solver_file=solver_file,
          centerline_file=centerline_file,rom_file=rom_file,flow_file=flow_file):
    cco_filename = options.filename
    if options.directory_constructor['create_main_folder']:
        os.mkdir(options.directory_constructor['main_folder'])
    cco_file = open(cco_filename,'w+')
    #setup_file = open(os.path.dirname(__file__)+'/setup.txt','r').read()
    if options.file_constructor['path']:
        #path_file = open(os.path.dirname(__file__)+'/paths.txt','r').read()
        path_file = path_file.format(options.file_constructor['gui'])
    else:
        path_file = ''
    if options.file_constructor['contour']:
        #contour_file = open(os.path.dirname(__file__)+'/contour.txt','r').readlines()
        tmp_contour_file = deepcopy(contour_file)
        tmp_contour_file[8] = tmp_contour_file[8].format(options.file_constructor['gui'])
        #print(contour_file[8])
        for i in range(len(tmp_contour_file)):
            tmp_contour_file[i] += "\n"
    else:
        contour_file = ''
    if options.file_constructor['solid']:
        #geometry_file = open(os.path.dirname(__file__)+'/geometry.txt','r').read()
        #geometry_file = geometry_file.format(options.geometry_options['number_samples'],
        #                                     options.geometry_options['use_distance_alignment_method'])
        #loft_file = open(os.path.dirname(__file__)+'/loft.txt','r').read()
        #loft_file = loft_file.format(options.loft_options['changes'],
        #                             options.loft_options['angle'])
        #solid_file = open(os.path.dirname(__file__)+'/solid.txt','r').read()
        #solid_file = solid_file.format(options.solid_options['minimum_face_cells'],
        #                               options.solid_options['hmin'],
        #                               options.solid_options['hmax'],
        #                               options.solid_options['face_edge_size'],
        #                               options.solid_options['face_edge_size'],
        #                               options.solid_options['angle'],
        #                               options.file_constructor['gui'],
        #                               options.name)
        loft_file = loft_file
        solid_file = solid_file.format(options.solid_options['num_caps'],options.solid_options['angle'],
                                       options.file_constructor['gui'],
                                       options.name)
    else:
        geometry_file = ''
        loft_file = ''
        solid_file = ''
    if options.file_constructor['mesh']:
        #mesh_file = open(os.path.dirname(__file__)+'/mesh.txt','r').read()
        mesh_file = mesh_file.format(options.mesh_options['global_edge_size'],
                                     options.mesh_options['surface_mesh_flag'],
                                     options.mesh_options['volume_mesh_flag'],
                                     options.mesh_options['no_merge'],
                                     options.mesh_options['optimization_level'],
                                     options.mesh_options['minimum_dihedral_angle'],
                                     options.file_constructor['gui'],
                                     options.name,
                                     options.name,
                                     options.directory_constructor['create_main_folder'],
                                     options.directory_constructor['main_folder'],
                                     options.directory_constructor['data_folder'],
                                     options.directory_constructor['mesh_complete_folder'],
                                     options.directory_constructor['mesh_surfaces_folder'],
                                     options.directory_constructor['centerline_folder'],
                                     options.directory_constructor['mesh_complete_folder'],
                                     options.directory_constructor['mesh_complete_folder'],
                                     options.directory_constructor['mesh_complete_folder'],
                                     options.directory_constructor['mesh_complete_folder'],
                                     options.directory_constructor['mesh_surfaces_folder'])
        #centerline_file = open(os.path.dirname(__file__)+'/centerline.txt','r').read()
        centerline_file = centerline_file.format(options.directory_constructor['centerline_folder'])
    else:
        mesh_files = ''
        centerline_file = ''
    #presolver_file = open(os.path.dirname(__file__)+'/presolver.txt','r').read()
    presolver_file = presolver_file.format(options.directory_constructor['data_folder'],
                                           options.presolver_constructor['name'],
                                           options.presolver_constructor['fluid_density'],
                                           options.presolver_constructor['fluid_viscosity'],
                                           options.presolver_constructor['initial_pressure'],
                                           options.presolver_constructor['initial_velocity'][0],
                                           options.presolver_constructor['initial_velocity'][1],
                                           options.presolver_constructor['initial_velocity'][2],
                                           options.presolver_constructor['bct_analytical_shape'],
                                           options.presolver_constructor['period'],
                                           options.presolver_constructor['bct_point_number'],
                                           options.presolver_constructor['fourier_mode'],
                                           options.presolver_constructor['inflow_file'],
                                           options.presolver_constructor['pressures'])
    #solver_file = open(os.path.dirname(__file__)+'/solver.txt','r').read()
    solver_file = solver_file.format(options.directory_constructor['data_folder'],
                                     options.solver_constructor['name'],
                                     options.solver_constructor['fluid_density'],
                                     options.solver_constructor['fluid_viscosity'],
                                     options.solver_constructor['number_timesteps'],
                                     options.solver_constructor['timestep_size'],
                                     options.solver_constructor['number_restarts'],
                                     options.solver_constructor['number_force_surfaces'],
                                     options.solver_constructor['surface_id_force_calc'],
                                     options.solver_constructor['force_calc_method'],
                                     options.solver_constructor['print_avg_solution'],
                                     options.solver_constructor['print_error_indicators'],
                                     options.solver_constructor['varying_time_from_file'],
                                     options.solver_constructor['step_construction'],
                                     options.solver_constructor['distal_resistance'],
                                     options.solver_constructor['pressure_coupling'],
                                     options.solver_constructor['backflow_stabilization'],
                                     options.solver_constructor['residual_control'],
                                     options.solver_constructor['residual_criteria'],
                                     options.solver_constructor['minimum_req_iter'],
                                     options.solver_constructor['svLS_type'],
                                     options.solver_constructor['num_krylov'],
                                     options.solver_constructor['num_solves_per_left'],
                                     options.solver_constructor['tolerance_momentum'],
                                     options.solver_constructor['tolerance_continuity'],
                                     options.solver_constructor['tolerance_svLS_NS'],
                                     options.solver_constructor['max_iter_NS'],
                                     options.solver_constructor['max_iter_momentum'],
                                     options.solver_constructor['max_iter_continuity'],
                                     options.solver_constructor['time_integration_rule'],
                                     options.solver_constructor['time_integration_rho'],
                                     options.solver_constructor['flow_advection_form'],
                                     options.solver_constructor['quadrature_interior'],
                                     options.solver_constructor['quadrature_boundary'],
                                     options.directory_constructor['data_folder'])
    flow_file = flow_file.format(options.directory_constructor['data_folder'],
                                 options.flow_constructor['inflow_file'],
                                 options.directory_constructor['data_folder'],
                                 options.flow_constructor['inflow_file'],
                                 options.flow_constructor['time'],
                                 options.flow_constructor['flow'])
    rom_files = 'import sv\n'
    if options.rom_constructor['0D']['create'] or options.rom_constructor['1D']['create']:
        rom_files += "os.mkdir('{}')\n".format(options.directory_constructor['rom_folder'])
    if options.rom_constructor['0D']['create']:
        rom_files += "os.mkdir('{}')\n".format(options.directory_constructor['rom_0D_folder'])
        rom0D = rom_file.format(options.name,
                                '0D',
                                options.directory_constructor['centerline_folder'],
                                options.directory_constructor['data_folder'],
                                options.rom_constructor['0D']['inflow_file'],
                                options.rom_constructor['0D']['time_step'],
                                options.rom_constructor['0D']['number_time_steps'],
                                options.directory_constructor['rom_0D_folder'],
                                options.rom_constructor['0D']['model_order'],
                                options.rom_constructor['0D']['model_order'],
                                options.directory_constructor['rom_0D_folder'],
                                options.directory_constructor['rom_0D_folder'],
                                options.rom_constructor['0D']['model_order'],
                                options.directory_constructor['rom_0D_folder'])
        rom_files += rom0D
    if options.rom_constructor['1D']['create']:
        rom_files += "os.mkdir('{}')\n".format(options.directory_constructor['rom_1D_folder'])
        rom1D = rom_file.format(options.name,
                                '1D',
                                options.directory_constructor['centerline_folder'],
                                options.directory_constructor['data_folder'],
                                options.rom_constructor['1D']['inflow_file'],
                                options.rom_constructor['1D']['time_step'],
                                options.rom_constructor['1D']['number_time_steps'],
                                options.directory_constructor['rom_1D_folder'],
                                options.rom_constructor['1D']['model_order'],
                                options.rom_constructor['1D']['model_order'],
                                options.directory_constructor['rom_1D_folder'],
                                options.directory_constructor['rom_1D_folder'],
                                options.rom_constructor['1D']['model_order'],
                                options.directory_constructor['rom_1D_folder'])
        rom_files += rom1D
    #postsolver_file = open('postsolver.txt','r').read()
    ############################################################
    # BUILDING SV PATHLINES
    ############################################################
    cco_file_text = ''
    cco_file_text += setup_file
    for idx, points in enumerate(paths):
        cco_file_text += path_file.format(idx,idx,points,idx,idx)
    ############################################################
    ############################################################

    ############################################################
    # BUILDING SV CONTOURS
    ############################################################
    cco_file_text += tmp_contour_file[0]
    cco_file_text += tmp_contour_file[1]
    for idx, path in enumerate(paths):
        cco_file_text += tmp_contour_file[2].format(idx)
        cco_file_text += tmp_contour_file[3].format(idx)
        contours = zip(paths[idx],radii[idx],normals[idx])
        for jdx,contour in enumerate(contours):
            cco_file_text += tmp_contour_file[4].format(idx,contour[1],contour[0],contour[2])
            cco_file_text += tmp_contour_file[5].format(idx,idx)
        cco_file_text += tmp_contour_file[6].format(idx)
        cco_file_text += tmp_contour_file[7].format(idx)
        cco_file_text += tmp_contour_file[8]
        cco_file_text += tmp_contour_file[9].format(idx,idx,idx)
    ############################################################
    ############################################################

    ############################################################
    # COMPUTING GEOMETRY
    ############################################################
    #cco_file_text += geometry_file
    ############################################################
    ############################################################

    ############################################################
    # LOFTING CONTOURS
    ############################################################
    cco_file_text += loft_file
    ############################################################
    ############################################################

    ############################################################
    # BUILDING SV SOLID
    ############################################################
    cco_file_text += solid_file
    ############################################################
    ############################################################

    ############################################################
    # BUILDING SV MESH
    ############################################################
    cco_file_text += mesh_file
    ############################################################
    ############################################################

    ############################################################
    # BUILDING SVPRE FILE
    ############################################################
    cco_file_text += presolver_file
    ############################################################
    ############################################################

    ############################################################
    # BUILDING SOLVER FILE
    ############################################################
    cco_file_text += solver_file
    ############################################################

    ############################################################
    # BUILD INFLOW FILES
    ############################################################
    cco_file_text += flow_file
    ############################################################

    ############################################################
    # EXTRACT CENTERLINE
    ############################################################
    cco_file_text += centerline_file
    ############################################################


    ############################################################
    # BUILDING ROM INPUT FILES
    ############################################################
    cco_file_text += rom_files
    ############################################################
    cco_file.writelines([cco_file_text])
    cco_file.close()



import numpy as np
import numba as nb

##################################################
# Adding Edge to Tree
##################################################

@nb.jit(nopython=True,cache=True,nogil=True)
def add_edge(data,proximal,distal,parent,flow,proximal_idx=-1.0,distal_idx=-1.0,
             left_child=-1.0,right_child=-1.0):
    edge = np.zeros(data.shape[1])
    # Set coordinates of new edge
    edge[0:3] = proximal
    edge[3:6] = distal
    # Normalized direction of added edge
    edge[12:15] = (distal - proximal) / np.linalg.norm(distal-proximal)
    # Get a random normal vector to the edge
    if edge[14] == -1.0:
        edge[6:9] = np.array([-1.0,0.0,0.0])
        edge[9:12] = np.array([0.0,-1.0,0.0])
    else:
        edge[6:9] = np.array([1.0-edge[12]**2/(1.0+edge[14]),
                              (-edge[12]*edge[13])/(1.0+edge[14]),
                              -edge[12]])
        edge[9:12] = np.array([(-edge[12]*edge[13])/(1.0+edge[14]),
                               1.0 - edge[13]**2/(1.0+edge[14]),
                               -edge[13]])
    # If this edge already had children then update them
    if left_child != -1.0:
        edge[15] = left_child.item()
    else:
        edge[15] = left_child
    if right_child != -1.0:
        edge[16] = right_child.item()
    else:
        edge[16] = right_child
    #edge[17] = parent
    # Determine node indicies for new edge
    if proximal_idx == -1.0:
        item1 = np.max(data[:,18])
        item2 = np.max(data[:,19])
        edge[18] = np.max(np.array([item1,item2])) + 1
    else:
        edge[18] = proximal_idx
    if distal_idx == -1.0:
        edge[19] = edge[18] + 1
    else:
        edge[19] = distal_idx
    # Length of new edge
    edge[20] = np.sqrt(np.sum(np.square(distal-proximal)))
    # index 21: radius, not determined yet
    edge[22] = flow
    # index 23: left bifurcation, not determined yet
    # index 24: right bifurcation, not determined yet
    # index 25: reduced resistance, not determined yet
    edge[26] = data[parent, 26].item() + 1
    if left_child != -1 and right_child != -1:
        edge[27] = data[parent, 27].item()
    # Actual edge
    edge[29] = -1
    # Index of edge
    edge[-1] = data.shape[0] - 1
    # Add edge to tree
    data[-1,:] = edge



solver="""if not terminating:
    solver_file = open('{}'+'/{}.inp','w+')
    solver_construction = 'Density: {}\\n'
    solver_construction += 'Viscosity: {}\\n\\n'
    solver_construction += 'Number of Timesteps: {}\\n'
    solver_construction += 'Time Step Size: {}\\n\\n'
    solver_construction += 'Number of Timesteps between Restarts: {}\\n'
    solver_construction += 'Number of Force Surfaces: {}\\n'
    solver_construction += 'Surface ID\\'s for Force Calculation: {}\\n'
    solver_construction += 'Force Calculation Method: {}\\n'
    solver_construction += 'Print Average Solution: {}\\n'
    solver_construction += 'Print Error Indicators: {}\\n\\n'
    solver_construction += 'Time Varying Boundary Conditions From File: {}\\n\\n'
    solver_construction += 'Step Construction: {}\\n\\n'
    num_res = len(mesher.get_model_face_ids()) - 2
    solver_construction += 'Number of Resistance Surfaces: {{}}\\n'.format(num_res)
    solver_construction += 'List of Resistance Surfaces:'
    for j in range(2,len(mesher.get_model_face_ids())+1):
        if j == skip:
            continue
        solver_construction += ' {{}}'.format(j)
    solver_construction += '\\n'
    solver_construction += 'Resistance Values:'
    faces = mesher.get_model_face_ids()
    outlet_res_total = {}
    for j in range(2,len(mesher.get_model_face_ids())+1):
        if j == skip:
            continue
        solver_construction += ' {{}}'.format(round(((1/resistance[faces[j-1]])*total_res)*outlet_res_total,2))
    solver_construction += '\\n\\n'
    solver_construction += 'Pressure Coupling: {}\\n'
    solver_construction += 'Number of Coupled Surfaces: {{}}\\n\\n'.format(num_res)
    solver_construction += 'Backflow Stabilization Coefficient: {}\\n'
    solver_construction += 'Residual Control: {}\\n'
    solver_construction += 'Residual Criteria: {}\\n'
    solver_construction += 'Minimum Required Iterations: {}\\n'
    solver_construction += 'svLS Type: {}\\n'
    solver_construction += 'Number of Krylov Vectors per GMRES Sweep: {}\\n'
    solver_construction += 'Number of Solves per Left-hand-side Formation: {}\\n'
    solver_construction += 'Tolerance on Momemtum Equations: {}\\n'
    solver_construction += 'Tolerance on Continuity Equations: {}\\n'
    solver_construction += 'Tolerance on svLS NS Solver: {}\\n'
    solver_construction += 'Maximum Number of Iterations for svLS NS Solver: {}\\n'
    solver_construction += 'Maximum Number of Iterations for svLS Momentum Loop: {}\\n'
    solver_construction += 'Maximum Number of Iterations for svLS Continuity Loop: {}\\n'
    solver_construction += 'Time Integration Rule: {}\\n'
    solver_construction += 'Time Integration Rho Infinity: {}\\n'
    solver_construction += 'Flow Advection Form: {}\\n'
    solver_construction += 'Quadrature Rule on Interior: {}\\n'
    solver_construction += 'Quadrature Rule on Boundary: {}\\n'
    solver_file.writelines([solver_construction])
    solver_file.close()
    numstart = open('{}'+'/numstart.dat','w+')
    numstart.writelines(['0'])
    numstart.close()

"""



mesh="""if not terminating:
    done = False
    min_edge = 0.001
    edge_size = {}
    attempt = 0
    while edge_size > min_edge and not done:
        try:
            faces = model.get_face_ids()
            mesher = meshing.create_mesher(meshing.Kernel.TETGEN)
            tet_options = meshing.TetGenOptions(edge_size,{},{})
            tet_options.no_merge = {}
            tet_options.optimization = {}
            tet_options.minimum_dihedral_angle = {}
            mesher.set_model(model)
            mesher.set_walls(walls)
            mesher.generate_mesh(tet_options)
            msh = mesher.get_mesh()
            done = True
        except:
            done = False
            edge_size = edge_size - 0.1*edge_size
            attempt += 1
    if {}:
        dmg.add_mesh('{}',msh,'{}')
    if {}:
        os.mkdir('{}')
    os.mkdir('{}')
    os.mkdir('{}')
    os.mkdir('{}')
    os.mkdir('{}')
    model.write('{}'+os.sep+'model_tmp','vtp')
    mesher.write_mesh('{}'+os.sep+'mesh-complete.mesh.vtu')
    mesh_out = modeling.PolyData()
    mesh_out.set_surface(mesher.get_surface())
    mesh_out.write('{}'+os.sep+'mesh-complete','exterior.vtp')
    mesh_out.set_surface(mesher.get_face_polydata(walls[0]))
    mesh_out.write('{}'+os.sep+'walls_combined','vtp')
    for face in mesher.get_model_face_ids():
        if face == walls[0]:
            continue
        mesh_out.set_surface(mesher.get_face_polydata(face))
        mesh_out.write('{}'+os.sep+'cap_{{}}'.format(face),'vtp')
"""



import numpy as np

from ...core.h_matrix import H
from ...core.utils import norm

def init_normals_not_given(lam,K00,K01,K11):
    H_matrix = H(K00,K01,K11,lam)
    n = K00.shape[0]
    d = K11.shape[0]//n
    w,v = np.linalg.eig(H_matrix)
    rw = w.real
    eh = np.argmin(rw)
    g0 = np.zeros((n,d))
    for i in range(n):
        #for j in range(d):
        #    g0[i,j] = np.real(v[eh,i+n*j])
        #mag = np.linalg.norm(g0[i,:])
        #for j in range(d):
        #    g0[i,j] = g0[i,j]/mag
        g0[i,0] = np.real(v[eh,i])
        g0[i,1] = np.real(v[eh,i+n])
        g0[i,2] = np.real(v[eh,i+n*2])
        norms = norm(g0[i,:])
        g0[i,0] = g0[i,0]/norms
        g0[i,1] = g0[i,1]/norms
        g0[i,2] = g0[i,2]/norms
    init_norms = np.zeros(n*(d-1))
    #for i in range(n):
    #    for j in range(d-1):
    #        if j == d - 1:
    #            if ic_norms[i,j] >= 0:
    #                init_norms[i*(d-1)+j] = np.arccos(g0[i,j]/np.linalg.norm(g0[i,j:]))
    #            else:
    #                init_norms[i*(d-1)+j] = 2*np.pi - np.arccos(g0[i,j]/np.linalg.norm(g0[i,j:]))
    #        else:
    #            init_norms[i*(d-1)+j] = np.arccos(g0[i,j]/np.linalg.norm(g0[i,j:]))
    for i in range(K00.shape[0]):
        init_norms[i*2] = np.arctan2((g0[i,0]**2+g0[i,1]**2)**(1/2),g0[i,2])
        init_norms[i*2+1] = np.arctan2(g0[i,1],g0[i,0])
    return init_norms



import numpy as np
import numba as nb

@nb.jit(nopython=True,cache=True,nogil=True)
def sphere_proximity(data,edge):
    C1 = (edge[0:3] + edge[3:6]) / 2
    C2 = (data[:,0:3] + data[:,3:6]) / 2
    r1 = np.sqrt(edge[21]**2+(edge[20]/2)**2)
    r2 = np.sqrt(data[:,21]**2+(data[:,20]/2)**2)
    D = np.sqrt(np.sum(np.square(C1 - C2), axis=1))
    return np.argwhere(D < (r1 + r2)).flatten()

@nb.jit(nopython=True,cache=True,nogil=True)
def sphere_point_query(point,distance,data):
    C = (data[:,0:3] + data[:,3:6]) / 2
    R = np.sqrt(data[:,21]**2+(data[:,20]/2)**2)
    D = np.sqrt(np.sum(np.square(C - point), axis=1))
    return np.argwhere(D<(distance+R)).flatten()

@nb.jit(nopython=True,nogil=True)
def sphere_proximity_testing(data,edge):
    C1 = (edge[0:3] + edge[3:6]) / 2
    C2 = (data[:,0:3] + data[:,3:6]) / 2
    r1 = np.sqrt(edge[21]**2+(edge[20]/2)**2)
    r2 = np.sqrt(data[:,21]**2+(data[:,20]/2)**2)
    D = np.sqrt(np.sum(np.square(C1 - C2), axis=1))
    return np.argwhere(D < (r1 + r2)).flatten()



import numpy as np

def init_normals_given(ic_norms):
    init_norms = np.zeros(ic_norms.shape[0]*(ic_norms.shape[1]-1))
    for i in range(ic_norms.shape[0]):
        #for j in range(ic_norms.shape[1]-1):
        #    if j == ic_norms.shape[1] - 1:
        #        if ic_norms[i,j] >= 0:
        #            init_norms[i*(ic_norms.shape[1]-1)+j] = np.arccos(ic_norms[i,j]/np.linalg.norm(ic_norms[i,j:]))
        #        else:
        #            init_norms[i*(ic_norms.shape[1]-1)+j] = 2*np.pi - np.arccos(ic_norms[i,j]/np.linalg.norm(ic_norms[i,j:]))
        #    else:
        #        init_norms[i*(ic_norms.shape[1]-1)+j] = np.arccos(ic_norms[i,j]/np.linalg.norm(ic_norms[i,j:]))
        init_norms[i*2] = np.arctan2((ic_norms[i,0]**2+ic_norms[i,1]**2)**(1/2),ic_norms[i,2])
        init_norms[i*2+1] = np.arctan2(ic_norms[i,1],ic_norms[i,0])
    return init_norms



import svcco
import pyvista as pv

sph = svcco.bumpy_sphere()
cube = pv.Cube().triangulate().subdivide(3)

s1 = svcco.surface()
s2 = svcco.surface()

s1.set_data(sph)
s2.set_data(3*cube.points,normals=cube.point_normals)

s1.solve()
s2.solve()

s1.build()
s2.build()
s1.subtract(s2)

t = svcco.tree()
t.set_boundary(s1)
t.parameters['Qterm'] *= 40
t.set_root()
t.n_add(100)
t.show(surface=True)



import numpy as np
import numba as nb

#@nb.jit(nopython=True,cache=True,nogil=True)
def sample_triad(p1,p2,p3):
    c = (p1+p2+p3)/3
    p12 = (p1+p2)/2
    p23 = (p2+p3)/2
    p31 = (p3+p1)/2
    return np.array([[p1,p12,c],[p12,p2,c],[p2,p23,c],[p23,p3,c],[p3,p31,c],[p31,p1,c]])

#@nb.jit(nopython=True,cache=True,nogil=True)
def triangle_subdivide(triangles,i=1):
    for u in range(i):
        subdivisions = []
        for triangle in triangles:
            divisions = sample_triad(triangle[0],triangle[1],triangle[2])
            for tri in divisions:
               subdivisions.append(tri)
        triangles = np.array(subdivisions)
    return triangles

#@nb.jit(nopython=True,cache=True,nogil=True)
def unique_triangle_points(triangles):
    a,b,c = triangles.shape
    pts = triangles.reshape(a*b,c)
    return np.unique(pts,axis=0)